# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# Unit Tests for Elements of a Model
#
# Test             Class to test the Model class
#

import json
import os
from os.path import abspath, dirname, join

currdir = dirname(abspath(__file__))
import pickle

from filecmp import cmp
import pyomo.common.unittest as unittest

from pyomo.common.dependencies import yaml_available
from pyomo.common.tempfiles import TempfileManager
import pyomo.core.expr as EXPR
from pyomo.environ import (
    RangeSet,
    ConcreteModel,
    Var,
    Param,
    Block,
    AbstractModel,
    Set,
    Constraint,
    Objective,
    value,
    sum_product,
    SolverFactory,
    VarList,
    ObjectiveList,
    ConstraintList,
    Model,
)
from pyomo.opt import check_available_solvers
from pyomo.opt.parallel.local import SolverManager_Serial

if yaml_available:
    import yaml

solvers = check_available_solvers('glpk')


class Test(unittest.TestCase):
    def tearDown(self):
        if os.path.exists("unknown.lp"):
            os.unlink("unknown.lp")
        TempfileManager.clear_tempfiles()

    def test_clone_concrete_model(self):
        def _populate(b, *args):
            b.A = RangeSet(1, 3)
            b.v = Var()
            b.vv = Var(m.A)
            b.p = Param()

        m = ConcreteModel()
        _populate(m)
        m.b = Block()
        _populate(m.b)
        m.b.c = Block()
        _populate(m.b.c)
        m.bb = Block(m.A, rule=_populate)

        n = m.clone()
        self.assertNotEqual(id(m), id(n))

        self.assertEqual(id(m), id(m.b.parent_block()))
        self.assertEqual(id(m), id(m.bb.parent_block()))
        self.assertEqual(id(m), id(m.bb[1].parent_block()))
        self.assertEqual(id(m), id(m.bb[2].parent_block()))
        self.assertEqual(id(m), id(m.bb[3].parent_block()))

        self.assertEqual(id(n), id(n.b.parent_block()))
        self.assertEqual(id(n), id(n.bb.parent_block()))
        self.assertEqual(id(n), id(n.bb[1].parent_block()))
        self.assertEqual(id(n), id(n.bb[2].parent_block()))
        self.assertEqual(id(n), id(n.bb[3].parent_block()))

        for x, y in ((m, n), (m.b, n.b), (m.b.c, n.b.c), (m.bb[2], n.bb[2])):
            self.assertNotEqual(id(x), id(y))
            self.assertNotEqual(id(x.parent_block()), id(x))
            self.assertNotEqual(id(y.parent_block()), id(y))

            self.assertEqual(id(x.A.parent_block()), id(x))
            self.assertEqual(id(x.v.parent_block()), id(x))
            self.assertEqual(id(x.vv.parent_block()), id(x))
            self.assertEqual(id(x.vv[1].parent_block()), id(x))
            self.assertEqual(id(x.p.parent_block()), id(x))

            self.assertEqual(id(y.A.parent_block()), id(y))
            self.assertEqual(id(y.v.parent_block()), id(y))
            self.assertEqual(id(y.vv.parent_block()), id(y))
            self.assertEqual(id(y.vv[1].parent_block()), id(y))
            self.assertEqual(id(y.p.parent_block()), id(y))

    def test_clone_abstract_model(self):
        def _populate(b, *args):
            b.A = RangeSet(1, 3)
            b.v = Var()
            b.vv = Var(m.A)
            b.p = Param()

        m = AbstractModel()
        _populate(m)
        m.b = Block()
        _populate(m.b)
        m.b.c = Block()
        _populate(m.b.c)
        m.bb = Block(m.A, rule=_populate)

        n = m.clone()
        self.assertNotEqual(id(m), id(n))

        self.assertEqual(id(m), id(m.b.parent_block()))
        self.assertEqual(id(m), id(m.bb.parent_block()))

        self.assertEqual(id(n), id(n.b.parent_block()))
        self.assertEqual(id(n), id(n.bb.parent_block()))

        for x, y in ((m, n), (m.b, n.b), (m.b.c, n.b.c)):
            self.assertNotEqual(id(x), id(y))
            self.assertNotEqual(id(x.parent_block()), id(x))
            self.assertNotEqual(id(y.parent_block()), id(y))

            self.assertEqual(id(x.A.parent_block()), id(x))
            self.assertEqual(id(x.v.parent_block()), id(x))
            self.assertEqual(id(x.vv.parent_block()), id(x))
            self.assertEqual(id(x.p.parent_block()), id(x))

            self.assertEqual(id(y.A.parent_block()), id(y))
            self.assertEqual(id(y.v.parent_block()), id(y))
            self.assertEqual(id(y.vv.parent_block()), id(y))
            self.assertEqual(id(y.p.parent_block()), id(y))

    def test_clear_attribute(self):
        # Test coverage of the _clear_attribute method
        model = ConcreteModel()
        obj = Set()
        model.A = obj
        self.assertEqual(model.A.local_name, "A")
        self.assertEqual(obj.local_name, "A")
        self.assertIs(obj, model.A)

        obj = Var()
        model.A = obj
        self.assertEqual(model.A.local_name, "A")
        self.assertEqual(obj.local_name, "A")
        self.assertIs(obj, model.A)

        obj = Param()
        model.A = obj
        self.assertEqual(model.A.local_name, "A")
        self.assertEqual(obj.local_name, "A")
        self.assertIs(obj, model.A)

        obj = Objective()
        model.A = obj
        self.assertEqual(model.A.local_name, "A")
        self.assertEqual(obj.local_name, "A")
        self.assertIs(obj, model.A)

        obj = Constraint()
        model.A = obj
        self.assertEqual(model.A.local_name, "A")
        self.assertEqual(obj.local_name, "A")
        self.assertIs(obj, model.A)

        obj = Set()
        model.A = obj
        self.assertEqual(model.A.local_name, "A")
        self.assertEqual(obj.local_name, "A")
        self.assertIs(obj, model.A)

    def test_set_attr(self):
        model = ConcreteModel()
        model.x = Param(mutable=True)
        model.x = 5
        self.assertEqual(value(model.x), 5)
        model.x = 6
        self.assertEqual(value(model.x), 6)
        model.x = None
        self.assertEqual(model.x._value, None)

    def test_write(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, bounds=(-1, 1))

        def obj_rule(model):
            return sum_product(model.x)

        model.obj = Objective(rule=obj_rule)
        model.write()

    def test_write2(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, bounds=(-1, 1))

        def obj_rule(model):
            return sum_product(model.x)

        model.obj = Objective(rule=obj_rule)

        def c_rule(model):
            return (1, model.x[1] + model.x[2], 2)

        model.c = Constraint(rule=c_rule)
        model.write()

    def test_write3(self):
        # Test that the summation works correctly, even though param 'w' has a default value
        model = ConcreteModel()
        model.J = RangeSet(1, 4)
        model.w = Param(model.J, default=4)
        model.x = Var(model.J, initialize=3)

        def obj_rule(instance):
            return sum_product(instance.w, instance.x)

        model.obj = Objective(rule=obj_rule)
        self.assertEqual(value(model.obj), 48)

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_solve1(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, bounds=(-1, 1))

        def obj_rule(model):
            return sum_product(model.x)

        model.obj = Objective(rule=obj_rule)

        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i * model.x[i]
            return expr == 0

        model.c = Constraint(rule=c_rule)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=True)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, "solve1.out"), format='json')
        with (
            open(join(currdir, "solve1.out"), 'r') as out,
            open(join(currdir, "solve1.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )

        #
        def d_rule(model):
            return model.x[1] >= 0

        model.d = Constraint(rule=d_rule)
        model.d.deactivate()
        results = opt.solve(model)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, "solve1x.out"), format='json')
        with (
            open(join(currdir, "solve1x.out"), 'r') as out,
            open(join(currdir, "solve1.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )
        #
        model.d.activate()
        results = opt.solve(model)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, "solve1a.out"), format='json')
        with (
            open(join(currdir, "solve1a.out"), 'r') as out,
            open(join(currdir, "solve1a.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )
        #
        model.d.deactivate()

        def e_rule(model, i):
            return model.x[i] >= 0

        model.e = Constraint(model.A, rule=e_rule)
        for i in model.A:
            model.e[i].deactivate()
        results = opt.solve(model)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, "solve1y.out"), format='json')
        with (
            open(join(currdir, "solve1y.out"), 'r') as out,
            open(join(currdir, "solve1.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )
        #
        model.e.activate()
        results = opt.solve(model)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, "solve1b.out"), format='json')
        with (
            open(join(currdir, "solve1b.out"), 'r') as out,
            open(join(currdir, "solve1b.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_store_to_skip_stale_vars(self):
        # test store_to() function with skip_stale_vars=True
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, bounds=(-1, 1))

        def obj_rule(model):
            return sum_product(model.x)

        model.obj = Objective(rule=obj_rule)

        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i * model.x[i]
            return expr == 0

        model.c = Constraint(rule=c_rule)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=True)
        model.x[1].fix()
        results = opt.solve(model, symbolic_solver_labels=True)
        model.solutions.store_to(results, skip_stale_vars=False)
        for index in model.A:
            self.assertIn(model.x[index].getname(), results.solution.variable.keys())
        model.solutions.store_to(results, skip_stale_vars=True)
        for index in model.A:
            if index == 1:
                self.assertNotIn(
                    model.x[index].getname(), results.solution.variable.keys()
                )
            else:
                self.assertIn(
                    model.x[index].getname(), results.solution.variable.keys()
                )

    def test_display(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, bounds=(-1, 1))

        def obj_rule(model):
            expr = 0
            for i in model.A:
                expr += model.x[i]
            return expr

        model.obj = Objective(rule=obj_rule)
        model.display(join(currdir, "solve3.out"))
        _out, _txt = join(currdir, "solve3.out"), join(currdir, "solve3.txt")
        self.assertTrue(cmp(_out, _txt), msg="Files %s and %s differ" % (_txt, _out))

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_solve4(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.x = Var(model.A, bounds=(-1, 1))

        def obj_rule(model):
            return sum_product(model.x)

        model.obj = Objective(rule=obj_rule)

        def c_rule(model):
            expr = 0
            for i in model.A:
                expr += i * model.x[i]
            return expr == 0

        model.c = Constraint(rule=c_rule)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=True)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve4.out'), format='json')
        with (
            open(join(currdir, "solve4.out"), 'r') as out,
            open(join(currdir, "solve1.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_solve6(self):
        #
        # Test that solution values have complete block names:
        #   b.obj
        #   b.x
        #
        model = ConcreteModel()
        model.y = Var(bounds=(-1, 1))
        model.b = Block()
        model.b.A = RangeSet(1, 4)
        model.b.x = Var(model.b.A, bounds=(-1, 1))

        def obj_rule(block):
            return sum_product(block.x)

        model.b.obj = Objective(rule=obj_rule)

        def c_rule(model):
            expr = model.y
            for i in model.b.A:
                expr += i * model.b.x[i]
            return expr == 0

        model.c = Constraint(rule=c_rule)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=True)
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve6.out'), format='json')
        with (
            open(join(currdir, "solve6.out"), 'r') as out,
            open(join(currdir, "solve6.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_solve7(self):
        #
        # Test that solution values are written with appropriate
        # quotations in results
        #
        model = ConcreteModel()
        model.y = Var(bounds=(-1, 1))
        model.A = RangeSet(1, 4)
        model.B = Set(initialize=['A B', 'C,D', 'E'])
        model.x = Var(model.A, model.B, bounds=(-1, 1))

        def obj_rule(model):
            return sum_product(model.x)

        model.obj = Objective(rule=obj_rule)

        def c_rule(model):
            expr = model.y
            for i in model.A:
                for j in model.B:
                    expr += i * model.x[i, j]
            return expr == 0

        model.c = Constraint(rule=c_rule)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=True)
        # model.display()
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve7.out'), format='json')
        with (
            open(join(currdir, "solve7.out"), 'r') as out,
            open(join(currdir, "solve7.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                json.load(txt), json.load(out), abstol=1e-4, allow_second_superset=True
            )

    def test_stats1(self):
        model = ConcreteModel()
        model.x = Var([1, 2])

        def obj_rule(model, i):
            return sum_product(model.x)

        model.obj = Objective([1, 2], rule=obj_rule)

        def c_rule(model, i):
            expr = 0
            for j in [1, 2]:
                expr += j * model.x[j]
            return expr == 0

        model.c = Constraint([1, 2], rule=c_rule)
        self.assertEqual(model.nvariables(), 2)
        self.assertEqual(model.nobjectives(), 2)
        self.assertEqual(model.nconstraints(), 2)

    def test_stats2(self):
        model = ConcreteModel()
        #
        model.x = Var([1, 2])

        def obj_rule(model, i):
            return sum_product(model.x)

        model.y = VarList()
        model.y.add()
        model.y.add()
        #
        model.obj = Objective([1, 2], rule=obj_rule)
        model.o = ObjectiveList()
        model.o.add(model.y[1])
        model.o.add(model.y[2])

        #
        def c_rule(model, i):
            expr = 0
            for j in [1, 2]:
                expr += j * model.x[j]
            return expr == 0

        model.c = Constraint([1, 2], rule=c_rule)
        model.C = ConstraintList()
        model.C.add(model.y[1] == 0)
        model.C.add(model.y[2] == 0)
        #
        self.assertEqual(model.nvariables(), 4)
        self.assertEqual(model.nobjectives(), 4)
        self.assertEqual(model.nconstraints(), 4)

    def test_stats3(self):
        model = ConcreteModel()
        model.x = Var([1, 2])

        def obj_rule(model, i):
            return sum_product(model.x)

        model.obj = Objective([1, 2], rule=obj_rule)

        def c_rule(model, i):
            expr = 0
            for j in [1, 2]:
                expr += j * model.x[j]
            return expr == 0

        model.c = Constraint([1, 2], rule=c_rule)
        #
        model.B = Block()
        model.B.x = Var([1, 2])
        model.B.o = ObjectiveList()
        model.B.o.add(model.B.x[1])
        model.B.o.add(model.B.x[2])
        model.B.c = ConstraintList()
        model.B.c.add(model.x[1] == 0)
        model.B.c.add(model.x[2] == 0)
        self.assertEqual(model.nvariables(), 4)
        self.assertEqual(model.nobjectives(), 4)
        self.assertEqual(model.nconstraints(), 4)

    def test_stats4(self):
        model = ConcreteModel()
        model.x = Var([1])

        model.B = Block()
        model.B.x = Var([1, 2, 3])
        model.B.o = ObjectiveList()
        model.B.o.add(model.B.x[1])
        model.B.c = ConstraintList()
        model.B.c.add(model.B.x[1] == 0)
        model.B.c.add(model.B.x[2] == 0)
        model.B.c.add(model.B.x[3] == 0)
        self.assertEqual(model.nvariables(), 4)
        self.assertEqual(model.nobjectives(), 1)
        self.assertEqual(model.nconstraints(), 3)
        model.clear()
        self.assertEqual(model.nvariables(), 0)
        self.assertEqual(model.nobjectives(), 0)
        self.assertEqual(model.nconstraints(), 0)

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_solve_with_pickle(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        self.assertEqual(len(model.solutions), 0)
        results = opt.solve(model, symbolic_solver_labels=True)
        self.assertEqual(len(model.solutions), 1)
        #
        self.assertEqual(model.solutions[0].gap, 0.0)
        # self.assertEqual(model.solutions[0].status, SolutionStatus.feasible)
        self.assertEqual(model.solutions[0].message, None)
        #
        buf = pickle.dumps(model)
        tmodel = pickle.loads(buf)
        self.assertEqual(len(tmodel.solutions), 1)
        self.assertEqual(tmodel.solutions[0].gap, 0.0)
        # self.assertEqual(tmodel.solutions[0].status, SolutionStatus.feasible)
        self.assertEqual(tmodel.solutions[0].message, None)

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    def test_solve_with_pickle_then_clone(self):
        # This tests github issue Pyomo-#65
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        self.assertEqual(len(model.solutions), 0)
        results = opt.solve(model, symbolic_solver_labels=True)
        self.assertEqual(len(model.solutions), 1)
        #
        self.assertEqual(model.solutions[0].gap, 0.0)
        # self.assertEqual(model.solutions[0].status, SolutionStatus.feasible)
        self.assertEqual(model.solutions[0].message, None)
        #
        buf = pickle.dumps(model)
        tmodel = pickle.loads(buf)
        self.assertEqual(len(tmodel.solutions), 1)
        self.assertEqual(tmodel.solutions[0].gap, 0.0)
        # self.assertEqual(tmodel.solutions[0].status, SolutionStatus.feasible)
        self.assertEqual(tmodel.solutions[0].message, None)
        self.assertIn(id(tmodel.b.obj), tmodel.solutions[0]._entry['objective'])
        self.assertIs(
            tmodel.b.obj, tmodel.solutions[0]._entry['objective'][id(tmodel.b.obj)][0]
        )

        inst = tmodel.clone()

        # make sure the clone has all the attributes
        self.assertTrue(hasattr(inst, 'A'))
        self.assertTrue(hasattr(inst, 'b'))
        self.assertTrue(hasattr(inst.b, 'x'))
        self.assertTrue(hasattr(inst.b, 'obj'))
        self.assertTrue(hasattr(inst, 'c'))
        # and that they were all copied
        self.assertIsNot(inst.A, tmodel.A)
        self.assertIsNot(inst.b, tmodel.b)
        self.assertIsNot(inst.b.x, tmodel.b.x)
        self.assertIsNot(inst.b.obj, tmodel.b.obj)
        self.assertIsNot(inst.c, tmodel.c)

        # Make sure the solution is on the new model
        self.assertTrue(hasattr(inst, 'solutions'))
        self.assertEqual(len(inst.solutions), 1)
        self.assertEqual(inst.solutions[0].gap, 0.0)
        # self.assertEqual(inst.solutions[0].status, SolutionStatus.feasible)
        self.assertEqual(inst.solutions[0].message, None)

        # Spot-check some components and make sure all the weakrefs in
        # the ModelSOlution got updated
        self.assertIn(id(inst.b.obj), inst.solutions[0]._entry['objective'])
        _obj = inst.solutions[0]._entry['objective'][id(inst.b.obj)]
        self.assertIs(_obj[0], inst.b.obj)

        for v in [1, 2, 3, 4]:
            self.assertIn(id(inst.b.x[v]), inst.solutions[0]._entry['variable'])
            _v = inst.solutions[0]._entry['variable'][id(inst.b.x[v])]
            self.assertIs(_v[0], inst.b.x[v])

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    @unittest.skipIf(not yaml_available, "YAML not available available")
    def test_solve_with_store1(self):
        # With symbolic solver labels
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=True)
        #
        results.write(filename=join(currdir, 'solve_with_store1.out'), format='yaml')
        with (
            open(join(currdir, "solve_with_store1.out"), 'r') as out,
            open(join(currdir, "solve_with_store1.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        model.solutions.store_to(results)
        #
        results.write(filename=join(currdir, 'solve_with_store2.out'), format='yaml')
        with (
            open(join(currdir, "solve_with_store2.out"), 'r') as out,
            open(join(currdir, "solve_with_store2.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        # Load results with string indices
        #
        tmodel = ConcreteModel()
        tmodel.A = RangeSet(1, 4)
        tmodel.b = Block()
        tmodel.b.x = Var(tmodel.A, bounds=(-1, 1))
        tmodel.b.obj = Objective(expr=sum_product(tmodel.b.x))
        tmodel.c = Constraint(expr=tmodel.b.x[1] >= 0)
        self.assertEqual(len(tmodel.solutions), 0)
        tmodel.solutions.load_from(results)
        self.assertEqual(len(tmodel.solutions), 1)

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    @unittest.skipIf(not yaml_available, "YAML not available available")
    def test_solve_with_store2(self):
        # Without symbolic solver labels
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        results = opt.solve(model, symbolic_solver_labels=False)
        #
        results.write(filename=join(currdir, 'solve_with_store1.out'), format='yaml')
        with (
            open(join(currdir, "solve_with_store1.out"), 'r') as out,
            open(join(currdir, "solve_with_store1.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        model.solutions.store_to(results)
        #
        results.write(filename=join(currdir, 'solve_with_store2.out'), format='yaml')
        with (
            open(join(currdir, "solve_with_store2.out"), 'r') as out,
            open(join(currdir, "solve_with_store2.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        # Load results with string indices
        #
        tmodel = ConcreteModel()
        tmodel.A = RangeSet(1, 4)
        tmodel.b = Block()
        tmodel.b.x = Var(tmodel.A, bounds=(-1, 1))
        tmodel.b.obj = Objective(expr=sum_product(tmodel.b.x))
        tmodel.c = Constraint(expr=tmodel.b.x[1] >= 0)
        self.assertEqual(len(tmodel.solutions), 0)
        tmodel.solutions.load_from(results)
        self.assertEqual(len(tmodel.solutions), 1)

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    @unittest.skipIf(not yaml_available, "YAML not available available")
    def test_solve_with_store2(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        results = opt.solve(model)
        #
        results.write(filename=join(currdir, 'solve_with_store3.out'), format='json')
        with (
            open(join(currdir, "solve_with_store3.out"), 'r') as out,
            open(join(currdir, "solve_with_store3.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve_with_store4.out'), format='json')
        with (
            open(join(currdir, "solve_with_store4.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        # Test that we can pickle the results object
        #
        buf = pickle.dumps(results)
        results_ = pickle.loads(buf)
        results.write(filename=join(currdir, 'solve_with_store4.out'), format='json')
        with (
            open(join(currdir, "solve_with_store4.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        # Load results with string indices
        #
        tmodel = ConcreteModel()
        tmodel.A = RangeSet(1, 3)
        tmodel.b = Block()
        tmodel.b.x = Var(tmodel.A, bounds=(-1, 1))
        tmodel.b.obj = Objective(expr=sum_product(tmodel.b.x))
        tmodel.c = Constraint(expr=tmodel.b.x[1] >= 0)
        self.assertEqual(len(tmodel.solutions), 0)
        tmodel.solutions.load_from(results, ignore_invalid_labels=True)
        self.assertEqual(len(tmodel.solutions), 1)

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    @unittest.skipIf(not yaml_available, "YAML not available available")
    def test_solve_with_store3(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        results = opt.solve(model)
        #
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve_with_store5.out'), format='json')
        with (
            open(join(currdir, "solve_with_store5.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        model.solutions.store_to(results, cuid=True)
        buf = pickle.dumps(results)
        results_ = pickle.loads(buf)
        model.solutions.load_from(results_)
        model.solutions.store_to(results_)
        results_.write(filename=join(currdir, 'solve_with_store6.out'), format='json')
        with (
            open(join(currdir, "solve_with_store6.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )
        #
        # Load results with string indices
        #
        tmodel = ConcreteModel()
        tmodel.A = RangeSet(1, 4)
        tmodel.b = Block()
        tmodel.b.x = Var(tmodel.A, bounds=(-1, 1))
        tmodel.b.obj = Objective(expr=sum_product(tmodel.b.x))
        tmodel.c = Constraint(expr=tmodel.b.x[1] >= 0)
        self.assertEqual(len(tmodel.solutions), 0)
        tmodel.solutions.load_from(results)
        self.assertEqual(len(tmodel.solutions), 1)
        tmodel.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve_with_store7.out'), format='json')
        with (
            open(join(currdir, "solve_with_store7.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    @unittest.skipIf(not yaml_available, "YAML not available available")
    def test_solve_with_store4(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)
        opt = SolverFactory('glpk')
        results = opt.solve(model, load_solutions=False)
        self.assertEqual(len(model.solutions), 0)
        self.assertEqual(len(results.solution), 1)
        model.solutions.load_from(results)
        self.assertEqual(len(model.solutions), 1)
        self.assertEqual(len(results.solution), 1)
        #
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve_with_store8.out'), format='json')
        with (
            open(join(currdir, "solve_with_store8.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )

    @unittest.skipIf('glpk' not in solvers, "glpk solver is not available")
    @unittest.skipIf(not yaml_available, "YAML not available available")
    def test_solve_with_store5(self):
        model = ConcreteModel()
        model.A = RangeSet(1, 4)
        model.b = Block()
        model.b.x = Var(model.A, bounds=(-1, 1))
        model.b.obj = Objective(expr=sum_product(model.b.x))
        model.c = Constraint(expr=model.b.x[1] >= 0)

        smanager = SolverManager_Serial()
        ah = smanager.queue(model, solver='glpk', load_solutions=False)
        results = smanager.wait_for(ah)
        self.assertEqual(len(model.solutions), 0)
        self.assertEqual(len(results.solution), 1)
        model.solutions.load_from(results)
        self.assertEqual(len(model.solutions), 1)
        self.assertEqual(len(results.solution), 1)
        #
        model.solutions.store_to(results)
        results.write(filename=join(currdir, 'solve_with_store8.out'), format='json')
        with (
            open(join(currdir, "solve_with_store8.out"), 'r') as out,
            open(join(currdir, "solve_with_store4.txt"), 'r') as txt,
        ):
            self.assertStructuredAlmostEqual(
                yaml.full_load(txt), yaml.full_load(out), allow_second_superset=True
            )

    def test_create_concrete_from_rule(self):
        def make(m):
            m.I = RangeSet(3)
            m.x = Var(m.I)
            m.c = Constraint(expr=sum(m.x[i] for i in m.I) >= 0)

        model = ConcreteModel(rule=make)
        self.assertEqual(
            [x.local_name for x in model.component_objects()], ['I', 'x', 'c']
        )
        self.assertEqual(len(list(EXPR.identify_variables(model.c.body))), 3)

    def test_create_abstract_from_rule(self):
        def make_invalid(m):
            m.I = RangeSet(3)
            m.x = Var(m.I)
            m.c = Constraint(expr=sum(m.x[i] for i in m.I) >= 0)

        def make(m):
            m.I = RangeSet(3)
            m.x = Var(m.I)

            def c(b):
                return sum(m.x[i] for i in m.I) >= 0

            m.c = Constraint(rule=c)

        with self.assertRaisesRegex(
            ValueError, r'x\[1\]: The component has not been constructed.'
        ):
            model = AbstractModel(rule=make_invalid)
            instance = model.create_instance()

        model = AbstractModel(rule=make)
        instance = model.create_instance()
        self.assertEqual([x.local_name for x in model.component_objects()], [])
        self.assertEqual(
            [x.local_name for x in instance.component_objects()], ['I', 'x', 'c']
        )
        self.assertEqual(len(list(EXPR.identify_variables(instance.c.body))), 3)

        model = AbstractModel(rule=make)
        model.y = Var()
        instance = model.create_instance()
        self.assertEqual(
            [x.local_name for x in instance.component_objects()], ['y', 'I', 'x', 'c']
        )
        self.assertEqual(len(list(EXPR.identify_variables(instance.c.body))), 3)

    def test_error_creating_model_baseclass(self):
        with self.assertRaisesRegex(
            TypeError,
            "Directly creating the 'Model' class is not allowed.  Please use the AbstractModel or ConcreteModel class instead.",
        ):
            m = Model()


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from typing import Sequence, Dict, Optional, Mapping

from pyomo.core.base.constraint import ConstraintData
from pyomo.core.base.var import VarData
from pyomo.core.staleflag import StaleFlagManager


class SolutionLoaderBase:
    """
    Base class for all future SolutionLoader classes.

    Intent of this class and its children is to load the solution back into the model.
    """

    def load_vars(self, vars_to_load: Optional[Sequence[VarData]] = None) -> None:
        """
        Load the solution of the primal variables into the value attribute of the variables.

        Parameters
        ----------
        vars_to_load: list
            The minimum set of variables whose solution should be loaded. If vars_to_load
            is None, then the solution to all primal variables will be loaded. Even if
            vars_to_load is specified, the values of other variables may also be
            loaded depending on the interface.
        """
        for var, val in self.get_primals(vars_to_load=vars_to_load).items():
            var.set_value(val, skip_validation=True)
        StaleFlagManager.mark_all_as_stale(delayed=True)

    def get_primals(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Returns a ComponentMap mapping variable to var value.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose solution value should be retrieved. If vars_to_load
            is None, then the values for all variables will be retrieved.

        Returns
        -------
        primals: ComponentMap
            Maps variables to solution values
        """
        raise NotImplementedError(
            f"Derived class {self.__class__.__name__} failed to implement required method 'get_primals'."
        )

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        """
        Returns a dictionary mapping constraint to dual value.

        Parameters
        ----------
        cons_to_load: list
            A list of the constraints whose duals should be retrieved. If cons_to_load
            is None, then the duals for all constraints will be retrieved.

        Returns
        -------
        duals: dict
            Maps constraints to dual values
        """
        raise NotImplementedError(f'{type(self)} does not support the get_duals method')

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        """
        Returns a ComponentMap mapping variable to reduced cost.

        Parameters
        ----------
        vars_to_load: list
            A list of the variables whose reduced cost should be retrieved. If vars_to_load
            is None, then the reduced costs for all variables will be loaded.

        Returns
        -------
        reduced_costs: ComponentMap
            Maps variables to reduced costs
        """
        raise NotImplementedError(
            f'{type(self)} does not support the get_reduced_costs method'
        )


class PersistentSolutionLoader(SolutionLoaderBase):
    """
    Loader for persistent solvers
    """

    def __init__(self, solver):
        self._solver = solver
        self._valid = True

    def _assert_solution_still_valid(self):
        if not self._valid:
            raise RuntimeError('The results in the solver are no longer valid.')

    def get_primals(self, vars_to_load=None):
        self._assert_solution_still_valid()
        return self._solver._get_primals(vars_to_load=vars_to_load)

    def get_duals(
        self, cons_to_load: Optional[Sequence[ConstraintData]] = None
    ) -> Dict[ConstraintData, float]:
        self._assert_solution_still_valid()
        return self._solver._get_duals(cons_to_load=cons_to_load)

    def get_reduced_costs(
        self, vars_to_load: Optional[Sequence[VarData]] = None
    ) -> Mapping[VarData, float]:
        self._assert_solution_still_valid()
        return self._solver._get_reduced_costs(vars_to_load=vars_to_load)

    def invalidate(self):
        self._valid = False



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import math
import copy
import re
import io
import pyomo.environ as pyo
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
from pyomo.core.expr import (
    NegationExpression,
    ProductExpression,
    DivisionExpression,
    PowExpression,
    AbsExpression,
    UnaryFunctionExpression,
    MonomialTermExpression,
    LinearExpression,
    SumExpression,
    EqualityExpression,
    InequalityExpression,
    RangedExpression,
    Expr_ifExpression,
    ExternalFunctionExpression,
)

from pyomo.core.expr.visitor import identify_components
from pyomo.core.expr.base import ExpressionBase
from pyomo.core.base.expression import ScalarExpression, ExpressionData
from pyomo.core.base.objective import ScalarObjective, ObjectiveData
import pyomo.core.kernel as kernel
from pyomo.core.expr.template_expr import (
    GetItemExpression,
    GetAttrExpression,
    TemplateSumExpression,
    IndexTemplate,
    Numeric_GetItemExpression,
    templatize_constraint,
    resolve_template,
    templatize_rule,
)
from pyomo.core.base.var import ScalarVar, VarData, IndexedVar
from pyomo.core.base.param import ParamData, ScalarParam, IndexedParam
from pyomo.core.base.set import SetData, SetOperator
from pyomo.core.base.constraint import ScalarConstraint, IndexedConstraint
from pyomo.common.collections.component_map import ComponentMap
from pyomo.common.collections.component_set import ComponentSet
from pyomo.core.expr.template_expr import (
    NPV_Numeric_GetItemExpression,
    NPV_Structural_GetItemExpression,
    Numeric_GetAttrExpression,
)
from pyomo.core.expr.numeric_expr import NPV_SumExpression, NPV_DivisionExpression
from pyomo.core.base.block import IndexedBlock

from pyomo.core.base.external import _PythonCallbackFunctionID
from pyomo.core.base.enums import SortComponents

from pyomo.core.base.block import BlockData

from pyomo.repn.util import ExprType

from pyomo.common import DeveloperError

_CONSTANT = ExprType.CONSTANT
_MONOMIAL = ExprType.MONOMIAL
_GENERAL = ExprType.GENERAL

from pyomo.common.errors import InfeasibleConstraintException

from pyomo.common.dependencies import numpy as np, numpy_available

set_operator_map = {
    '|': r' \cup ',
    '&': r' \cap ',
    '*': r' \times ',
    '-': r' \setminus ',
    '^': r' \triangle ',
}

latex_reals = r'\mathds{R}'
latex_integers = r'\mathds{Z}'

domainMap = {
    'Reals': latex_reals,
    'PositiveReals': latex_reals + '_{> 0}',
    'NonPositiveReals': latex_reals + '_{\\leq 0}',
    'NegativeReals': latex_reals + '_{< 0}',
    'NonNegativeReals': latex_reals + '_{\\geq 0}',
    'Integers': latex_integers,
    'PositiveIntegers': latex_integers + '_{> 0}',
    'NonPositiveIntegers': latex_integers + '_{\\leq 0}',
    'NegativeIntegers': latex_integers + '_{< 0}',
    'NonNegativeIntegers': latex_integers + '_{\\geq 0}',
    'Boolean': '\\left\\{ \\text{True} , \\text{False} \\right \\}',
    'Binary': '\\left\\{ 0 , 1 \\right \\}',
    # 'Any': None,
    # 'AnyWithNone': None,
    'EmptySet': '\\varnothing',
    'UnitInterval': latex_reals,
    'PercentFraction': latex_reals,
    # 'RealInterval' :        None    ,
    # 'IntegerInterval' :     None    ,
}


def decoder(num, base):
    if int(num) != abs(num):
        # Requiring an integer is nice, but not strictly necessary;
        # the algorithm works for floating point
        raise ValueError("num should be a nonnegative integer")
    if int(base) != abs(base) or not base:
        raise ValueError("base should be a positive integer")
    ans = []
    while 1:
        ans.append(num % base)
        num //= base
        if not num:
            return list(reversed(ans))


def indexCorrector(ixs, base):
    for i in range(0, len(ixs)):
        ix = ixs[i]
        if i + 1 < len(ixs):
            if ixs[i + 1] == 0:
                ixs[i] -= 1
                ixs[i + 1] = base
                if ixs[i] == 0:
                    ixs = indexCorrector(ixs, base)
    return ixs


def alphabetStringGenerator(num):
    alphabet = ['.', 'i', 'j', 'k', 'm', 'n', 'p', 'q', 'r']

    ixs = decoder(num + 1, len(alphabet) - 1)
    pstr = ''
    ixs = indexCorrector(ixs, len(alphabet) - 1)
    for i in range(0, len(ixs)):
        ix = ixs[i]
        pstr += alphabet[ix]
        pstr = pstr.replace('.', '')
    return pstr


def templatize_expression(expr):
    expr, indices = templatize_rule(expr.parent_block(), expr._rule, expr.index_set())
    return (expr, indices)


def templatize_passthrough(con):
    return (con, [])


def precedenceChecker(node, arg1, arg2=None):
    childPrecedence = []
    for a in node.args:
        if hasattr(a, 'PRECEDENCE'):
            if a.PRECEDENCE is None:
                childPrecedence.append(-1)
            else:
                childPrecedence.append(a.PRECEDENCE)
        else:
            childPrecedence.append(-1)

    if hasattr(node, 'PRECEDENCE'):
        precedence = node.PRECEDENCE
    else:
        # Should never hit this
        raise DeveloperError(
            'This error should never be thrown, node does not have a precedence.  Report to developers'
        )

    if childPrecedence[0] > precedence:
        arg1 = ' \\left( ' + arg1 + ' \\right) '

    if arg2 is not None:
        if childPrecedence[1] > precedence:
            arg2 = ' \\left( ' + arg2 + ' \\right) '

    return arg1, arg2


def handle_negation_node(visitor, node, arg1):
    arg1, tsh = precedenceChecker(node, arg1)
    return '-' + arg1


def handle_product_node(visitor, node, arg1, arg2):
    arg1, arg2 = precedenceChecker(node, arg1, arg2)
    return ' '.join([arg1, arg2])


def handle_pow_node(visitor, node, arg1, arg2):
    arg1, arg2 = precedenceChecker(node, arg1, arg2)
    return "%s^{%s}" % (arg1, arg2)


def handle_division_node(visitor, node, arg1, arg2):
    return '\\frac{%s}{%s}' % (arg1, arg2)


def handle_abs_node(visitor, node, arg1):
    return ' \\left| ' + arg1 + ' \\right| '


def handle_unary_node(visitor, node, arg1):
    fcn_handle = node.getname()
    if fcn_handle == 'log10':
        fcn_handle = 'log_{10}'

    if fcn_handle == 'sqrt':
        return '\\sqrt { ' + arg1 + ' }'
    else:
        return '\\' + fcn_handle + ' \\left( ' + arg1 + ' \\right) '


def handle_equality_node(visitor, node, arg1, arg2):
    return arg1 + ' = ' + arg2


def handle_inequality_node(visitor, node, arg1, arg2):
    return arg1 + ' \\leq ' + arg2


def handle_var_node(visitor, node):
    return visitor.variableMap[node]


def handle_num_node(visitor, node):
    if isinstance(node, float):
        if node.is_integer():
            node = int(node)
    return str(node)


def handle_sumExpression_node(visitor, node, *args):
    rstr = args[0]
    for i in range(1, len(args)):
        if args[i][0] == '-':
            rstr += ' - ' + args[i][1:]
        else:
            rstr += ' + ' + args[i]
    return rstr


def handle_monomialTermExpression_node(visitor, node, arg1, arg2):
    if arg1 == '1':
        return arg2
    elif arg1 == '-1':
        return '-' + arg2
    else:
        return arg1 + ' ' + arg2


def handle_named_expression_node(visitor, node, arg1):
    # needed to preserve consistency with the exitNode function call
    # prevents the need to type check in the exitNode function
    return arg1


def handle_ranged_inequality_node(visitor, node, arg1, arg2, arg3):
    return arg1 + ' \\leq ' + arg2 + ' \\leq ' + arg3


def handle_exprif_node(visitor, node, arg1, arg2, arg3):
    return 'f_{\\text{exprIf}}(' + arg1 + ',' + arg2 + ',' + arg3 + ')'

    ## Could be handled in the future using cases or similar

    ## Raises not implemented error
    # raise NotImplementedError('Expr_if objects not supported by the Latex Printer')

    ## Puts cases in a bracketed matrix
    # pstr = ''
    # pstr += '\\begin{Bmatrix} '
    # pstr += arg2 + ' , & ' + arg1 + '\\\\ '
    # pstr += arg3 + ' , & \\text{otherwise}' + '\\\\ '
    # pstr += '\\end{Bmatrix}'
    # return pstr


def handle_external_function_node(visitor, node, *args):
    pstr = ''
    visitor.externalFunctionCounter += 1
    pstr += 'f\\_' + str(visitor.externalFunctionCounter) + '('
    for i in range(0, len(args) - 1):
        pstr += args[i]
        if i <= len(args) - 3:
            pstr += ','
        else:
            pstr += ')'
    return pstr


def handle_functionID_node(visitor, node, *args):
    # seems to just be a placeholder empty wrapper object
    return ''


def handle_indexTemplate_node(visitor, node, *args):
    if node._set in visitor.setMap:
        # already detected set, do nothing
        pass
    else:
        visitor.setMap[node._set] = 'SET%d' % (len(visitor.setMap) + 1)

    return '__I_PLACEHOLDER_8675309_GROUP_%s_%s_%s__' % (
        node._group,
        node._id,
        visitor.setMap[node._set],
    )


def handle_numericGetItemExpression_node(visitor, node, *args):
    joinedName = args[0]

    pstr = ''
    pstr += joinedName + '_{'
    for i in range(1, len(args)):
        pstr += args[i]
        if i <= len(args) - 2:
            pstr += ','
        else:
            pstr += '}'
    return pstr


def handle_templateSumExpression_node(visitor, node, *args):
    pstr = ''
    for i in range(0, len(node._iters)):
        pstr += '\\sum_{__S_PLACEHOLDER_8675309_GROUP_%s_%s_%s__} ' % (
            node._iters[i][0]._group,
            ','.join(str(it._id) for it in node._iters[i]),
            visitor.setMap[node._iters[i][0]._set],
        )

    pstr += args[0]

    return pstr


def handle_param_node(visitor, node):
    return visitor.parameterMap[node]


def handle_str_node(visitor, node):
    return "\\mathtt{'" + node.replace('_', '\\_') + "'}"


def handle_npv_structuralGetItemExpression_node(visitor, node, *args):
    joinedName = args[0]

    pstr = ''
    pstr += joinedName + '['
    for i in range(1, len(args)):
        pstr += args[i]
        if i <= len(args) - 2:
            pstr += ','
        else:
            pstr += ']'
    return pstr


def handle_indexedBlock_node(visitor, node, *args):
    return str(node)


def handle_numericGetAttrExpression_node(visitor, node, *args):
    return args[0] + '.' + args[1]


class _LatexVisitor(StreamBasedExpressionVisitor):
    def __init__(self):
        super().__init__()
        self.externalFunctionCounter = 0

        self._operator_handles = {
            ScalarVar: handle_var_node,
            int: handle_num_node,
            float: handle_num_node,
            NegationExpression: handle_negation_node,
            ProductExpression: handle_product_node,
            DivisionExpression: handle_division_node,
            PowExpression: handle_pow_node,
            AbsExpression: handle_abs_node,
            UnaryFunctionExpression: handle_unary_node,
            Expr_ifExpression: handle_exprif_node,
            EqualityExpression: handle_equality_node,
            InequalityExpression: handle_inequality_node,
            RangedExpression: handle_ranged_inequality_node,
            ExpressionData: handle_named_expression_node,
            ScalarExpression: handle_named_expression_node,
            kernel.expression.expression: handle_named_expression_node,
            kernel.expression.noclone: handle_named_expression_node,
            ObjectiveData: handle_named_expression_node,
            VarData: handle_var_node,
            ScalarObjective: handle_named_expression_node,
            kernel.objective.objective: handle_named_expression_node,
            ExternalFunctionExpression: handle_external_function_node,
            _PythonCallbackFunctionID: handle_functionID_node,
            LinearExpression: handle_sumExpression_node,
            SumExpression: handle_sumExpression_node,
            MonomialTermExpression: handle_monomialTermExpression_node,
            IndexedVar: handle_var_node,
            IndexTemplate: handle_indexTemplate_node,
            Numeric_GetItemExpression: handle_numericGetItemExpression_node,
            TemplateSumExpression: handle_templateSumExpression_node,
            ScalarParam: handle_param_node,
            ParamData: handle_param_node,
            IndexedParam: handle_param_node,
            NPV_Numeric_GetItemExpression: handle_numericGetItemExpression_node,
            IndexedBlock: handle_indexedBlock_node,
            NPV_Structural_GetItemExpression: handle_npv_structuralGetItemExpression_node,
            str: handle_str_node,
            Numeric_GetAttrExpression: handle_numericGetAttrExpression_node,
            NPV_SumExpression: handle_sumExpression_node,
            NPV_DivisionExpression: handle_division_node,
        }
        if numpy_available:
            self._operator_handles[np.float64] = handle_num_node

    def exitNode(self, node, data):
        try:
            return self._operator_handles[node.__class__](self, node, *data)
        except:
            raise DeveloperError(
                'Latex printer encountered an error when processing type %s, contact the developers'
                % (node.__class__)
            )


def analyze_variable(vr):
    domainName = vr.domain.name
    varBounds = vr.bounds
    lowerBoundValue = varBounds[0]
    upperBoundValue = varBounds[1]

    if domainName in ['Reals', 'Integers']:
        if lowerBoundValue is not None:
            lowerBound = str(lowerBoundValue) + ' \\leq '
        else:
            lowerBound = ''

        if upperBoundValue is not None:
            upperBound = ' \\leq ' + str(upperBoundValue)
        else:
            upperBound = ''

    elif domainName in ['PositiveReals', 'PositiveIntegers']:
        if lowerBoundValue > 0:
            lowerBound = str(lowerBoundValue) + ' \\leq '
        else:
            lowerBound = ' 0 < '

        if upperBoundValue is not None:
            if upperBoundValue <= 0:
                raise InfeasibleConstraintException(
                    'Formulation is infeasible due to bounds on variable %s' % (vr.name)
                )
            else:
                upperBound = ' \\leq ' + str(upperBoundValue)
        else:
            upperBound = ''

    elif domainName in ['NonPositiveReals', 'NonPositiveIntegers']:
        if lowerBoundValue is not None:
            if lowerBoundValue > 0:
                raise InfeasibleConstraintException(
                    'Formulation is infeasible due to bounds on variable %s' % (vr.name)
                )
            elif lowerBoundValue == 0:
                lowerBound = ' 0 = '
            else:
                lowerBound = str(lowerBoundValue) + ' \\leq '
        else:
            lowerBound = ''

        if upperBoundValue >= 0:
            upperBound = ' \\leq 0 '
        else:
            upperBound = ' \\leq ' + str(upperBoundValue)

    elif domainName in ['NegativeReals', 'NegativeIntegers']:
        if lowerBoundValue is not None:
            if lowerBoundValue >= 0:
                raise InfeasibleConstraintException(
                    'Formulation is infeasible due to bounds on variable %s' % (vr.name)
                )
            else:
                lowerBound = str(lowerBoundValue) + ' \\leq '
        else:
            lowerBound = ''

        if upperBoundValue >= 0:
            upperBound = ' < 0 '
        else:
            upperBound = ' \\leq ' + str(upperBoundValue)

    elif domainName in ['NonNegativeReals', 'NonNegativeIntegers']:
        if lowerBoundValue > 0:
            lowerBound = str(lowerBoundValue) + ' \\leq '
        else:
            lowerBound = ' 0 \\leq '

        if upperBoundValue is not None:
            if upperBoundValue < 0:
                raise InfeasibleConstraintException(
                    'Formulation is infeasible due to bounds on variable %s' % (vr.name)
                )
            elif upperBoundValue == 0:
                upperBound = ' = 0 '
            else:
                upperBound = ' \\leq ' + str(upperBoundValue)
        else:
            upperBound = ''

    elif domainName in ['Boolean', 'Binary', 'Any', 'AnyWithNone', 'EmptySet']:
        lowerBound = ''
        upperBound = ''

    elif domainName in ['UnitInterval', 'PercentFraction']:
        if lowerBoundValue > 1:
            raise InfeasibleConstraintException(
                'Formulation is infeasible due to bounds on variable %s' % (vr.name)
            )
        elif lowerBoundValue == 1:
            lowerBound = ' = 1 '
        elif lowerBoundValue > 0:
            lowerBound = str(lowerBoundValue) + ' \\leq '
        else:
            lowerBound = ' 0 \\leq '

        if upperBoundValue < 0:
            raise InfeasibleConstraintException(
                'Formulation is infeasible due to bounds on variable %s' % (vr.name)
            )
        elif upperBoundValue == 0:
            upperBound = ' = 0 '
        elif upperBoundValue < 1:
            upperBound = ' \\leq ' + str(upperBoundValue)
        else:
            upperBound = ' \\leq 1 '

    else:
        raise NotImplementedError(
            'Invalid domain encountered, will be supported in a future update'
        )

    varBoundData = {
        'variable': vr,
        'lowerBound': lowerBound,
        'upperBound': upperBound,
        'domainName': domainName,
        'domainLatex': domainMap[domainName],
    }

    return varBoundData


def multiple_replace(pstr, rep_dict):
    pattern = re.compile("|".join(rep_dict.keys()), flags=re.DOTALL)
    return pattern.sub(lambda x: rep_dict[x.group(0)], pstr)


def latex_printer(
    pyomo_component,
    latex_component_map=None,
    ostream=None,
    use_equation_environment=False,
    explicit_set_summation=False,
    throw_templatization_error=False,
):
    """This function produces a string that can be rendered as LaTeX

    Prints a Pyomo component (Block, Model, Objective, Constraint, or Expression) to a LaTeX compatible string

    Parameters
    ----------
    pyomo_component: BlockData or Model or Objective or Constraint or Expression
        The Pyomo component to be printed

    latex_component_map: pyomo.common.collections.component_map.ComponentMap
        A map keyed by Pyomo component, values become the LaTeX representation in
        the printer

    ostream: io.TextIOWrapper or io.StringIO or str
        The object to print the LaTeX string to.  Can be an open file object,
        string I/O object, or a string for a filename to write to

    use_equation_environment: bool
        If False, the equation/aligned construction is used to create a single
         LaTeX equation.  If True, then the align environment is used in LaTeX and
         each constraint and objective will be given an individual equation number

    explicit_set_summation: bool
        If False, all sums will be done over 'index in set' or similar.  If True,
        sums will be done over 'i=1' to 'N' or similar if the set is a continuous
        set

    throw_templatization_error: bool
        Option to throw an error on templatization failure rather than
        printing each constraint individually, useful for very large models


    Returns
    -------
    str
        A LaTeX string of the pyomo_component

    """

    # Various setup things

    # is Single implies Objective, constraint, or expression
    # these objects require a slight modification of behavior
    # isSingle==False means a model or block

    use_short_descriptors = True

    # Cody's backdoor because he got outvoted
    if latex_component_map is not None:
        if 'use_short_descriptors' in latex_component_map:
            if latex_component_map['use_short_descriptors'] == False:
                use_short_descriptors = False

    if latex_component_map is None:
        latex_component_map = ComponentMap()
        existing_components = ComponentSet()
    else:
        existing_components = ComponentSet(latex_component_map)

    isSingle = False

    if isinstance(pyomo_component, pyo.Objective):
        objectives = [pyomo_component]
        constraints = []
        expressions = []
        templatize_fcn = templatize_constraint
        use_equation_environment = True
        isSingle = True

    elif isinstance(pyomo_component, pyo.Constraint):
        objectives = []
        constraints = [pyomo_component]
        expressions = []
        templatize_fcn = templatize_constraint
        use_equation_environment = True
        isSingle = True

    elif isinstance(pyomo_component, pyo.Expression):
        objectives = []
        constraints = []
        expressions = [pyomo_component]
        templatize_fcn = templatize_expression
        use_equation_environment = True
        isSingle = True

    elif isinstance(pyomo_component, (ExpressionBase, pyo.Var)):
        objectives = []
        constraints = []
        expressions = [pyomo_component]
        templatize_fcn = templatize_passthrough
        use_equation_environment = True
        isSingle = True

    elif isinstance(pyomo_component, BlockData):
        objectives = [
            obj
            for obj in pyomo_component.component_data_objects(
                pyo.Objective,
                descend_into=True,
                active=True,
                sort=SortComponents.deterministic,
            )
        ]
        constraints = [
            con
            for con in pyomo_component.component_objects(
                pyo.Constraint,
                descend_into=True,
                active=True,
                sort=SortComponents.deterministic,
            )
        ]
        expressions = []
        templatize_fcn = templatize_constraint

    else:
        raise ValueError(
            "Invalid type %s passed into the latex printer"
            % (str(type(pyomo_component)))
        )

    if isSingle:
        temp_comp, temp_indexes = templatize_fcn(pyomo_component)
        variableList = []
        for v in identify_components(temp_comp, [ScalarVar, VarData, IndexedVar]):
            if isinstance(v, VarData):
                v_write = v.parent_component()
                if v_write not in ComponentSet(variableList):
                    variableList.append(v_write)
            else:
                if v not in ComponentSet(variableList):
                    variableList.append(v)

        parameterList = []
        for p in identify_components(temp_comp, [ScalarParam, ParamData, IndexedParam]):
            if isinstance(p, ParamData):
                p_write = p.parent_component()
                if p_write not in ComponentSet(parameterList):
                    parameterList.append(p_write)
            else:
                if p not in ComponentSet(parameterList):
                    parameterList.append(p)

        # Will grab the sets as the expression is walked
        setList = []

    else:
        variableList = [
            vr
            for vr in pyomo_component.component_objects(
                pyo.Var,
                descend_into=True,
                active=True,
                sort=SortComponents.deterministic,
            )
        ]

        parameterList = [
            pm
            for pm in pyomo_component.component_objects(
                pyo.Param,
                descend_into=True,
                active=True,
                sort=SortComponents.deterministic,
            )
        ]

        setList = [
            st
            for st in pyomo_component.component_objects(
                pyo.Set,
                descend_into=True,
                active=True,
                sort=SortComponents.deterministic,
            )
        ]

    descriptorDict = {}
    if use_short_descriptors:
        descriptorDict['minimize'] = '\\min'
        descriptorDict['maximize'] = '\\max'
        descriptorDict['subject to'] = '\\text{s.t.}'
        descriptorDict['with bounds'] = '\\text{w.b.}'
    else:
        descriptorDict['minimize'] = '\\text{minimize}'
        descriptorDict['maximize'] = '\\text{maximize}'
        descriptorDict['subject to'] = '\\text{subject to}'
        descriptorDict['with bounds'] = '\\text{with bounds}'

    # In the case where just a single expression is passed, add this to the constraint list for printing
    constraints = constraints + expressions

    # Declare a visitor/walker
    visitor = _LatexVisitor()

    variableMap = ComponentMap()
    vrIdx = 0
    for vr in variableList:
        vrIdx += 1
        if isinstance(vr, ScalarVar):
            variableMap[vr] = 'x_' + str(vrIdx) + '_'
        elif isinstance(vr, IndexedVar):
            variableMap[vr] = 'x_' + str(vrIdx) + '_'
            for sd in vr.index_set().data():
                vrIdx += 1
                variableMap[vr[sd]] = 'x_' + str(vrIdx) + '_'
        else:
            raise DeveloperError(
                'Variable is not a variable.  Should not happen.  Contact developers'
            )
    visitor.variableMap = variableMap

    parameterMap = ComponentMap()
    pmIdx = 0
    for vr in parameterList:
        pmIdx += 1
        if isinstance(vr, ScalarParam):
            parameterMap[vr] = 'p_' + str(pmIdx) + '_'
        elif isinstance(vr, IndexedParam):
            parameterMap[vr] = 'p_' + str(pmIdx) + '_'
            for sd in vr.index_set().data():
                pmIdx += 1
                parameterMap[vr[sd]] = 'p_' + str(pmIdx) + '_'
        else:
            raise DeveloperError(
                'Parameter is not a parameter.  Should not happen.  Contact developers'
            )
    visitor.parameterMap = parameterMap

    setMap = ComponentMap()
    for i in range(0, len(setList)):
        st = setList[i]
        setMap[st] = 'SET' + str(i + 1)
    visitor.setMap = setMap

    # starts building the output string
    pstr = ''
    if not use_equation_environment:
        pstr += '\\begin{align} \n'
        tbSpc = 4
        trailingAligner = '& '
    else:
        pstr += '\\begin{equation} \n'
        if not isSingle:
            pstr += '    \\begin{aligned} \n'
            tbSpc = 8
        else:
            tbSpc = 4
        trailingAligner = ''

    # Iterate over the objectives and print
    for obj in objectives:
        try:
            obj_template, obj_indices = templatize_fcn(obj)
        except:
            if throw_templatization_error:
                raise RuntimeError(
                    "An objective named '%s' has been constructed that cannot be templatized"
                    % (obj.__str__())
                )
            else:
                obj_template = obj

        if obj.sense == pyo.minimize:  # or == 1
            pstr += ' ' * tbSpc + '& %s \n' % (descriptorDict['minimize'])
        else:
            pstr += ' ' * tbSpc + '& %s \n' % (descriptorDict['maximize'])

        pstr += ' ' * tbSpc + '& & %s %s' % (
            visitor.walk_expression(obj_template),
            trailingAligner,
        )
        if not use_equation_environment:
            pstr += '\\label{obj:' + pyomo_component.name + '_' + obj.name + '} '
        if not isSingle:
            pstr += '\\\\ \n'
        else:
            pstr += '\n'

    # Iterate over the constraints
    if len(constraints) > 0:
        # only print this if printing a full formulation
        if not isSingle:
            pstr += ' ' * tbSpc + '& %s \n' % (descriptorDict['subject to'])

        # first constraint needs different alignment because of the 'subject to':
        # & minimize   & & [Objective]
        # & subject to & & [Constraint 1]
        # &            & & [Constraint 2]
        # &            & & [Constraint N]

        # The double '& &' renders better for some reason

        for i, con in enumerate(constraints):
            if not isSingle:
                if i == 0:
                    algn = '& &'
                else:
                    algn = '&&&'
            else:
                algn = ''

            if not isSingle:
                tail = '\\\\ \n'
            else:
                tail = '\n'

            # grab the constraint and templatize
            try:
                con_template, indices = templatize_fcn(con)
                con_template_list = [con_template]
            except:
                if throw_templatization_error:
                    raise RuntimeError(
                        "A constraint named '%s' has been constructed that cannot be templatized"
                        % (con.__str__())
                    )
                else:
                    con_template_list = [c.expr for c in con.values()]
                    indices = []

            for con_template in con_template_list:
                # Walk the constraint
                conLine = (
                    ' ' * tbSpc
                    + algn
                    + ' %s %s'
                    % (visitor.walk_expression(con_template), trailingAligner)
                )

                # setMap = visitor.setMap
                # Multiple constraints are generated using a set
                if len(indices) > 0:
                    conLine += ' \\qquad \\forall'

                    _bygroups = {}
                    for idx in indices:
                        _bygroups.setdefault(idx._group, []).append(idx)
                    for _group, idxs in _bygroups.items():
                        if idxs[0]._set in visitor.setMap:
                            # already detected set, do nothing
                            pass
                        else:
                            visitor.setMap[idxs[0]._set] = 'SET%d' % (
                                len(visitor.setMap) + 1
                            )

                        idxTag = ','.join(
                            '__I_PLACEHOLDER_8675309_GROUP_%s_%s_%s__'
                            % (idx._group, idx._id, visitor.setMap[idx._set])
                            for idx in idxs
                        )

                        setTag = '__S_PLACEHOLDER_8675309_GROUP_%s_%s_%s__' % (
                            indices[0]._group,
                            ','.join(str(it._id) for it in idxs),
                            visitor.setMap[indices[0]._set],
                        )

                        conLine += ' %s \\in %s ' % (idxTag, setTag)
                pstr += conLine

                # Add labels as needed
                if not use_equation_environment:
                    pstr += (
                        '\\label{con:' + pyomo_component.name + '_' + con.name + '} '
                    )

                pstr += tail

    # Print bounds and sets
    if not isSingle:
        varBoundData = []
        for i in range(0, len(variableList)):
            vr = variableList[i]
            if isinstance(vr, ScalarVar):
                varBoundDataEntry = analyze_variable(vr)
                varBoundData.append(varBoundDataEntry)
            elif isinstance(vr, IndexedVar):
                varBoundData_indexedVar = []
                setData = vr.index_set().data()
                for sd in setData:
                    varBoundDataEntry = analyze_variable(vr[sd])
                    varBoundData_indexedVar.append(varBoundDataEntry)
                globIndexedVariables = True
                for j in range(0, len(varBoundData_indexedVar) - 1):
                    chks = []
                    chks.append(
                        varBoundData_indexedVar[j]['lowerBound']
                        == varBoundData_indexedVar[j + 1]['lowerBound']
                    )
                    chks.append(
                        varBoundData_indexedVar[j]['upperBound']
                        == varBoundData_indexedVar[j + 1]['upperBound']
                    )
                    chks.append(
                        varBoundData_indexedVar[j]['domainName']
                        == varBoundData_indexedVar[j + 1]['domainName']
                    )
                    if not all(chks):
                        globIndexedVariables = False
                        break
                if globIndexedVariables:
                    varBoundData.append(
                        {
                            'variable': vr,
                            'lowerBound': varBoundData_indexedVar[0]['lowerBound'],
                            'upperBound': varBoundData_indexedVar[0]['upperBound'],
                            'domainName': varBoundData_indexedVar[0]['domainName'],
                            'domainLatex': varBoundData_indexedVar[0]['domainLatex'],
                        }
                    )
                else:
                    varBoundData += varBoundData_indexedVar
            else:
                raise DeveloperError(
                    'Variable is not a variable.  Should not happen.  Contact developers'
                )

        # print the accumulated data to the string
        bstr = ''
        appendBoundString = False
        useThreeAlgn = False
        for i, vbd in enumerate(varBoundData):
            if (
                vbd['lowerBound'] == ''
                and vbd['upperBound'] == ''
                and vbd['domainName'] == 'Reals'
            ):
                # unbounded all real, do not print
                if i == len(varBoundData) - 1:
                    bstr = bstr[0:-4]
            else:
                if not useThreeAlgn:
                    algn = '& &'
                    useThreeAlgn = True
                else:
                    algn = '&&&'

                if use_equation_environment:
                    conLabel = ''
                else:
                    conLabel = (
                        ' \\label{con:'
                        + pyomo_component.name
                        + '_'
                        + variableMap[vbd['variable']]
                        + '_bound'
                        + '} '
                    )

                appendBoundString = True
                coreString = (
                    vbd['lowerBound']
                    + variableMap[vbd['variable']]
                    + vbd['upperBound']
                    + ' '
                    + trailingAligner
                    + '\\qquad \\in '
                    + vbd['domainLatex']
                    + conLabel
                )
                bstr += ' ' * tbSpc + algn + ' %s' % (coreString)
                if i <= len(varBoundData) - 2:
                    bstr += '\\\\ \n'
                else:
                    bstr += '\n'

        if appendBoundString:
            pstr += ' ' * tbSpc + '& %s \n' % (descriptorDict['with bounds'])
            pstr += bstr + '\n'
        else:
            pstr = pstr[0:-4] + '\n'

    # close off the print string
    if not use_equation_environment:
        pstr += '\\end{align} \n'
    else:
        if not isSingle:
            pstr += '    \\end{aligned} \n'
            pstr += '    \\label{%s} \n' % (pyomo_component.name)
        pstr += '\\end{equation} \n'

    setMap = visitor.setMap
    setMap_inverse = {vl: ky for ky, vl in setMap.items()}

    def generate_set_name(st, lcm):
        if st in lcm:
            return lcm[st][0]
        if st.parent_block().component(st.name) is st:
            return st.name.replace('_', r'\_')
        if isinstance(st, SetOperator):
            return set_operator_map[st._operator.strip()].join(
                generate_set_name(s, lcm) for s in st.subsets(False)
            )
        else:
            return str(st).replace('_', r'\_').replace('{', r'\{').replace('}', r'\}')

    # Handling the iterator indices
    defaultSetLatexNames = ComponentMap()
    for ky in setMap:
        defaultSetLatexNames[ky] = generate_set_name(ky, latex_component_map)

    latexLines = pstr.split('\n')
    for jj in range(0, len(latexLines)):
        groupMap = {}
        uniqueSets = []
        ln = latexLines[jj]
        # only modify if there is a placeholder in the line
        if "PLACEHOLDER_8675309_GROUP_" in ln:
            splitLatex = ln.split('__')
            # Find the unique combinations of group numbers and set names
            for word in splitLatex:
                if "PLACEHOLDER_8675309_GROUP_" in word:
                    ifo = word.split("PLACEHOLDER_8675309_GROUP_")[1]
                    gpNum, idNum, stName = ifo.split('_')
                    if gpNum not in groupMap:
                        groupMap[gpNum] = [stName]
                    if stName not in ComponentSet(uniqueSets):
                        uniqueSets.append(stName)

            # Determine if the set is continuous
            setInfo = dict(
                zip(
                    uniqueSets,
                    [{'continuous': False} for i in range(0, len(uniqueSets))],
                )
            )

            for ky, vl in setInfo.items():
                ix = int(ky[3:]) - 1
                setInfo[ky]['setObject'] = setMap_inverse[ky]  # setList[ix]
                setInfo[ky]['setRegEx'] = (
                    r'__S_PLACEHOLDER_8675309_GROUP_([0-9]+)_([0-9,]+)_%s__' % (ky,)
                )
                # setInfo[ky]['idxRegEx'] = r'__I_PLACEHOLDER_8675309_GROUP_[0-9*]_%s__'%(ky)

            if explicit_set_summation:
                for ky, vl in setInfo.items():
                    st = vl['setObject']
                    stData = st.data()
                    stCont = True
                    for ii in range(0, len(stData)):
                        if ii + stData[0] != stData[ii]:
                            stCont = False
                            break
                    setInfo[ky]['continuous'] = stCont

            # replace the sets
            for ky, vl in setInfo.items():
                # if the set is continuous and the flag has been set
                if explicit_set_summation and setInfo[ky]['continuous']:
                    st = setInfo[ky]['setObject']
                    stData = st.data()
                    bgn = stData[0]
                    ed = stData[-1]

                    replacement = (
                        r'sum_{ __I_PLACEHOLDER_8675309_GROUP_\1_\2_%s__ = %d }^{%d}'
                        % (ky, bgn, ed)
                    )
                    ln = re.sub(
                        'sum_{' + setInfo[ky]['setRegEx'] + '}', replacement, ln
                    )
                else:
                    # if the set is not continuous or the flag has not been set
                    for _grp, _id in re.findall(
                        'sum_{' + setInfo[ky]['setRegEx'] + '}', ln
                    ):
                        set_placeholder = '__S_PLACEHOLDER_8675309_GROUP_%s_%s_%s__' % (
                            _grp,
                            _id,
                            ky,
                        )
                        i_placeholder = ','.join(
                            '__I_PLACEHOLDER_8675309_GROUP_%s_%s_%s__' % (_grp, _, ky)
                            for _ in _id.split(',')
                        )
                        replacement = r'sum_{ %s \in %s  }' % (
                            i_placeholder,
                            set_placeholder,
                        )
                        ln = ln.replace('sum_{' + set_placeholder + '}', replacement)

                replacement = repr(defaultSetLatexNames[setInfo[ky]['setObject']])[1:-1]
                ln = re.sub(setInfo[ky]['setRegEx'], replacement, ln)

            # groupNumbers = re.findall(r'__I_PLACEHOLDER_8675309_GROUP_([0-9*])_SET[0-9]*__',ln)
            setNumbers = re.findall(
                r'__I_PLACEHOLDER_8675309_GROUP_[0-9]+_[0-9]+_SET([0-9]+)__', ln
            )
            groupIdSetTuples = re.findall(
                r'__I_PLACEHOLDER_8675309_GROUP_([0-9]+)_([0-9]+)_SET([0-9]+)__', ln
            )

            groupInfo = {}
            for vl in setNumbers:
                groupInfo['SET' + vl] = {
                    'setObject': setInfo['SET' + vl]['setObject'],
                    'indices': [],
                }

            for _gp, _id, _set in groupIdSetTuples:
                if (_gp, _id) not in groupInfo['SET' + _set]['indices']:
                    groupInfo['SET' + _set]['indices'].append((_gp, _id))

            def get_index_names(st, lcm):
                if st in lcm:
                    return lcm[st][1]
                elif isinstance(st, SetOperator):
                    return sum(
                        (get_index_names(s, lcm) for s in st.subsets(False)), start=[]
                    )
                elif st.dimen is not None:
                    return [None] * st.dimen
                else:
                    return [Ellipsis]

            indexCounter = 0
            for ky, vl in groupInfo.items():
                indexNames = get_index_names(vl['setObject'], latex_component_map)
                nonNone = list(filter(None, indexNames))
                if nonNone:
                    if len(nonNone) < len(vl['indices']):
                        raise ValueError(
                            'Insufficient number of indices provided to the '
                            'overwrite dictionary for set %s (expected %s, but got %s)'
                            % (vl['setObject'].name, len(vl['indices']), indexNames)
                        )
                else:
                    indexNames = []
                    for i in vl['indices']:
                        indexNames.append(alphabetStringGenerator(indexCounter))
                        indexCounter += 1
                for i in range(0, len(vl['indices'])):
                    ln = ln.replace(
                        '__I_PLACEHOLDER_8675309_GROUP_%s_%s_%s__'
                        % (*vl['indices'][i], ky),
                        indexNames[i],
                    )
        latexLines[jj] = ln

    pstr = '\n'.join(latexLines)

    new_variableMap = ComponentMap()
    for i, vr in enumerate(variableList):
        if isinstance(vr, ScalarVar):
            new_variableMap[vr] = vr.name
        elif isinstance(vr, IndexedVar):
            new_variableMap[vr] = vr.name
            for sd in vr.index_set().data():
                sdString = str(sd)
                if sdString[0] == '(':
                    sdString = sdString[1:]
                if sdString[-1] == ')':
                    sdString = sdString[0:-1]
                new_variableMap[vr[sd]] = vr[sd].name
        else:
            raise DeveloperError(
                'Variable is not a variable.  Should not happen.  Contact developers'
            )

    new_parameterMap = ComponentMap()
    for i, pm in enumerate(parameterList):
        pm = parameterList[i]
        if isinstance(pm, ScalarParam):
            new_parameterMap[pm] = pm.name
        elif isinstance(pm, IndexedParam):
            new_parameterMap[pm] = pm.name
            for sd in pm.index_set().data():
                sdString = str(sd)
                if sdString[0] == '(':
                    sdString = sdString[1:]
                if sdString[-1] == ')':
                    sdString = sdString[0:-1]
                new_parameterMap[pm[sd]] = str(pm[sd])  # .name
        else:
            raise DeveloperError(
                'Parameter is not a parameter.  Should not happen.  Contact developers'
            )

    for ky, vl in new_variableMap.items():
        if ky not in latex_component_map:
            latex_component_map[ky] = vl
    for ky, vl in new_parameterMap.items():
        if ky not in latex_component_map:
            latex_component_map[ky] = vl

    rep_dict = {}
    for ky in reversed(list(latex_component_map)):
        if isinstance(ky, (pyo.Var, VarData)):
            overwrite_value = latex_component_map[ky]
            if ky not in existing_components:
                overwrite_value = overwrite_value.replace('_', '\\_')
            rep_dict[variableMap[ky]] = overwrite_value
        elif isinstance(ky, (pyo.Param, ParamData)):
            overwrite_value = latex_component_map[ky]
            if ky not in existing_components:
                overwrite_value = overwrite_value.replace('_', '\\_')
            rep_dict[parameterMap[ky]] = overwrite_value
        elif isinstance(ky, SetData):
            # already handled
            pass
        elif isinstance(ky, (float, int)):
            # happens when immutable parameters are used, do nothing
            pass
        else:
            raise ValueError(
                'The latex_component_map object has a key of invalid type: %s'
                % (str(ky))
            )

    label_rep_dict = copy.deepcopy(rep_dict)
    for ky, vl in label_rep_dict.items():
        label_rep_dict[ky] = vl.replace('{', '').replace('}', '').replace('\\', '')

    splitLines = pstr.split('\n')
    for i in range(0, len(splitLines)):
        if use_equation_environment:
            splitLines[i] = multiple_replace(splitLines[i], rep_dict)
        else:
            if '\\label{' in splitLines[i]:
                epr, lbl = splitLines[i].split('\\label{')
                epr = multiple_replace(epr, rep_dict)
                # rep_dict[ky] = vl.replace('_', '\\_')
                lbl = multiple_replace(lbl, label_rep_dict)
                splitLines[i] = epr + '\\label{' + lbl

    pstr = '\n'.join(splitLines)

    pattern = r'_{([^{]*)}_{([^{]*)}'
    replacement = r'_{\1_{\2}}'
    pstr = re.sub(pattern, replacement, pstr)

    pattern = r'_(.)_{([^}]*)}'
    replacement = r'_{\1_{\2}}'
    pstr = re.sub(pattern, replacement, pstr)

    splitLines = pstr.split('\n')
    finalLines = []
    for sl in splitLines:
        if sl != '':
            finalLines.append(sl)

    pstr = '\n'.join(finalLines)

    if ostream is not None:
        fstr = ''
        fstr += '\\documentclass{article} \n'
        fstr += '\\usepackage{amsmath} \n'
        fstr += '\\usepackage{amssymb} \n'
        fstr += '\\usepackage{dsfont} \n'
        fstr += '\\usepackage[paperheight=11in, paperwidth=8.5in, left=1in, right=1in, top=1in, bottom=1in]{geometry} \n'
        fstr += '\\allowdisplaybreaks \n'
        fstr += '\\begin{document} \n'
        fstr += '\\normalsize \n'
        fstr += pstr + '\n'
        fstr += '\\end{document} \n'

        # optional write to output file
        if isinstance(ostream, (io.TextIOWrapper, io.StringIO)):
            ostream.write(fstr)
        elif isinstance(ostream, str):
            f = open(ostream, 'w')
            f.write(fstr)
            f.close()
        else:
            raise ValueError(
                'Invalid type %s encountered when parsing the ostream.  Must be a StringIO, FileIO, or valid filename string'
            )

    # return the latex string
    return pstr



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
#
# This module was originally developed as part of the IDAES PSE Framework
#
# Institute for the Design of Advanced Energy Systems Process Systems
# Engineering Framework (IDAES PSE Framework) Copyright (c) 2018-2019, by the
# software owners: The Regents of the University of California, through
# Lawrence Berkeley National Laboratory,  National Technology & Engineering
# Solutions of Sandia, LLC, Carnegie Mellon University, West Virginia
# University Research Corporation, et al. All rights reserved.
#
# This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
A simple GUI viewer/editor for Pyomo models.
"""

__author__ = "John Eslick"

import logging
import os

from pyomo.common.fileutils import this_file_dir
from pyomo.common.flags import building_documentation

import pyomo.contrib.viewer.qt as myqt
import pyomo.environ as pyo

_log = logging.getLogger(__name__)


# This lets the file be imported when the Qt UI is not available (or
# when building docs), but you won't be able to use it
class _ModelSelectUI:
    pass


class _ModelSelect:
    pass


# Note that the classes loaded here have signatures that are not
# parsable by Sphinx, so we won't attempt to import them if we are
# building the API documentation.
if not building_documentation():
    mypath = this_file_dir()
    try:
        _ModelSelectUI, _ModelSelect = myqt.uic.loadUiType(
            os.path.join(mypath, "model_select.ui")
        )
    except:
        pass


class ModelSelect(_ModelSelect, _ModelSelectUI):
    def __init__(self, parent, ui_data):
        super().__init__(parent)
        self.setupUi(self)
        self.ui_data = ui_data
        self.closeButton.clicked.connect(self.close)
        self.selectButton.clicked.connect(self.select_model)

    def select_model(self):
        items = self.tableWidget.selectedItems()
        if len(items) == 0:
            return
        self.ui_data.model_var_name_in_main = self.models[items[0].row()][1]
        self.ui_data.model = self.models[items[0].row()][0]
        self.close()

    def update_models(self):
        import __main__

        s = dir(__main__)
        keys = []
        for k in s:
            if isinstance(getattr(__main__, k), pyo.Block):
                keys.append(k)
        self.tableWidget.clearContents()
        self.tableWidget.setRowCount(len(keys))
        self.models = []
        for row, k in enumerate(sorted(keys)):
            model = getattr(__main__, k)
            item = myqt.QTableWidgetItem()
            item.setText(k)
            self.tableWidget.setItem(row, 0, item)
            item = myqt.QTableWidgetItem()
            try:
                item.setText(model.name)
            except:
                item.setText("None")
            self.tableWidget.setItem(row, 1, item)
            item = myqt.QTableWidgetItem()
            item.setText(str(type(model)))
            self.tableWidget.setItem(row, 2, item)
            self.models.append((model, k))



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections.abc import Callable
from typing import Any, Protocol

from pyomo.contrib.solver.solvers.knitro.typing import (
    Callback,
    CallbackFunction,
    CallbackRequest,
    CallbackResult,
    Function,
)

CallbackHandlerFunction = Callable[[CallbackRequest, CallbackResult], int]


class CallbackHandler(Protocol):
    _function: Function

    def func(self, req: CallbackRequest, res: CallbackResult) -> int: ...
    def grad(self, req: CallbackRequest, res: CallbackResult) -> int: ...
    def hess(self, req: CallbackRequest, res: CallbackResult) -> int: ...

    def expand(self) -> Callback:
        procs = (self.func, self.grad, self.hess)
        return Callback(*map(self._expand, procs))

    @staticmethod
    def _expand(proc: CallbackHandlerFunction) -> CallbackFunction:
        def _expanded(
            kc: Any,
            cb: Any,
            req: CallbackRequest,
            res: CallbackResult,
            user_data: Any = None,
        ) -> int:
            return proc(req, res)

        return _expanded


class ObjectiveCallbackHandler(CallbackHandler):
    def __init__(self, function: Function) -> None:
        self._function = function

    def func(self, req: CallbackRequest, res: CallbackResult) -> int:
        res.obj = self._function.evaluate(req.x)
        return 0

    def grad(self, req: CallbackRequest, res: CallbackResult) -> int:
        res.objGrad[:] = self._function.gradient(req.x)
        return 0

    def hess(self, req: CallbackRequest, res: CallbackResult) -> int:
        res.hess[:] = self._function.hessian(req.x, req.sigma)
        return 0


class ConstraintCallbackHandler(CallbackHandler):
    i: int

    def __init__(self, i: int, function: Function) -> None:
        self.i = i
        self._function = function

    def func(self, req: CallbackRequest, res: CallbackResult) -> int:
        res.c[:] = [self._function.evaluate(req.x)]
        return 0

    def grad(self, req: CallbackRequest, res: CallbackResult) -> int:
        res.jac[:] = self._function.gradient(req.x)
        return 0

    def hess(self, req: CallbackRequest, res: CallbackResult) -> int:
        res.hess[:] = self._function.hessian(req.x, req.lambda_[self.i])
        return 0


def build_callback_handler(function: Function, idx: int | None) -> CallbackHandler:
    if idx is None:
        return ObjectiveCallbackHandler(function)
    return ConstraintCallbackHandler(idx, function)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import logging
import platform
from pyomo.common import Library
from pyomo.common.deprecation import deprecated

logger = logging.getLogger('pyomo.common')


@deprecated(
    "Use of get_gsl is deprecated and NO LONGER FUNCTIONS as of February 9, 2023. ",
    version='6.5.0',
)
def get_gsl(downloader):
    logger.info("As of February 9, 2023, AMPL GSL can no longer be downloaded \
        through download-extensions. Visit https://portal.ampl.com/ \
        to download the AMPL GSL binaries.")


def find_GSL():
    # FIXME: the GSL interface is currently broken in PyPy:
    if platform.python_implementation().lower().startswith('pypy'):
        return None
    return Library('amplgsl.dll').path()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.environ as pyo
from pyomo.contrib.pynumero.interfaces.pyomo_nlp import PyomoNLP
from pyomo.contrib.pynumero.sparse import BlockMatrix, BlockVector
from scipy.sparse import identity
from scipy.sparse.linalg import spsolve
import numpy as np


def create_model(eta1, eta2):
    model = pyo.ConcreteModel()
    # variables
    model.x1 = pyo.Var(initialize=0.15)
    model.x2 = pyo.Var(initialize=0.15)
    model.x3 = pyo.Var(initialize=0.0)
    # parameters
    model.eta1 = pyo.Var()
    model.eta2 = pyo.Var()

    model.nominal_eta1 = pyo.Param(initialize=eta1, mutable=True)
    model.nominal_eta2 = pyo.Param(initialize=eta2, mutable=True)

    # constraints + objective
    model.const1 = pyo.Constraint(
        expr=6 * model.x1 + 3 * model.x2 + 2 * model.x3 - model.eta1 == 0
    )
    model.const2 = pyo.Constraint(
        expr=model.eta2 * model.x1 + model.x2 - model.x3 - 1 == 0
    )
    model.cost = pyo.Objective(expr=model.x1**2 + model.x2**2 + model.x3**2)
    model.consteta1 = pyo.Constraint(expr=model.eta1 == model.nominal_eta1)
    model.consteta2 = pyo.Constraint(expr=model.eta2 == model.nominal_eta2)

    return model


def compute_init_lam(nlp, x=None, lam_max=1e3):
    if x is None:
        x = nlp.init_primals()
    else:
        assert x.size == nlp.n_primals()
    nlp.set_primals(x)

    assert (
        nlp.n_ineq_constraints() == 0
    ), "only supported for equality constrained nlps for now"

    nx = nlp.n_primals()
    nc = nlp.n_constraints()

    # create Jacobian
    jac = nlp.evaluate_jacobian()

    # create gradient of objective
    df = nlp.evaluate_grad_objective()

    # create KKT system
    kkt = BlockMatrix(2, 2)
    kkt.set_block(0, 0, identity(nx))
    kkt.set_block(1, 0, jac)
    kkt.set_block(0, 1, jac.transpose())

    zeros = np.zeros(nc)
    rhs = BlockVector(2)
    rhs.set_block(0, -df)
    rhs.set_block(1, zeros)

    flat_kkt = kkt.tocoo().tocsc()
    flat_rhs = rhs.flatten()

    sol = spsolve(flat_kkt, flat_rhs)
    return sol[nlp.n_primals() : nlp.n_primals() + nlp.n_constraints()]


def main():
    m = create_model(4.5, 1.0)
    opt = pyo.SolverFactory('ipopt')
    results = opt.solve(m, tee=True)

    nlp = PyomoNLP(m)
    x = nlp.init_primals()
    y = compute_init_lam(nlp, x=x)
    nlp.set_primals(x)
    nlp.set_duals(y)

    J = nlp.extract_submatrix_jacobian(
        pyomo_variables=[m.x1, m.x2, m.x3], pyomo_constraints=[m.const1, m.const2]
    )
    H = nlp.extract_submatrix_hessian_lag(
        pyomo_variables_rows=[m.x1, m.x2, m.x3], pyomo_variables_cols=[m.x1, m.x2, m.x3]
    )

    M = BlockMatrix(2, 2)
    M.set_block(0, 0, H)
    M.set_block(1, 0, J)
    M.set_block(0, 1, J.transpose())

    Np = BlockMatrix(2, 1)
    Np.set_block(
        0,
        0,
        nlp.extract_submatrix_hessian_lag(
            pyomo_variables_rows=[m.x1, m.x2, m.x3],
            pyomo_variables_cols=[m.eta1, m.eta2],
        ),
    )
    Np.set_block(
        1,
        0,
        nlp.extract_submatrix_jacobian(
            pyomo_variables=[m.eta1, m.eta2], pyomo_constraints=[m.const1, m.const2]
        ),
    )

    ds = spsolve(M.tocsc(), -Np.tocsc())

    print("ds:\n", ds.todense())
    #################################################################

    p0 = np.array([pyo.value(m.nominal_eta1), pyo.value(m.nominal_eta2)])
    p = np.array([4.45, 1.05])
    dp = p - p0
    dx = ds.dot(dp)[0:3]
    x_indices = nlp.get_primal_indices([m.x1, m.x2, m.x3])
    x_names = np.array(nlp.primals_names())
    new_x_sens = x[x_indices] + dx
    print("dp:", dp)
    print("dx:", dx)
    print("Variable names: \n", x_names[x_indices])
    print("Sensitivity based x:\n", new_x_sens)

    #################################################################
    m = create_model(4.45, 1.05)
    opt = pyo.SolverFactory('ipopt')
    results = opt.solve(m, tee=False)
    nlp = PyomoNLP(m)
    new_x = nlp.init_primals()[nlp.get_primal_indices([m.x1, m.x2, m.x3])]
    print("NLP based x:\n", new_x)

    return new_x_sens, new_x


if __name__ == '__main__':
    main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.common.dependencies import attempt_import
from pyomo.common.dependencies import numpy as np, numpy_available

from pyomo.contrib.solver.tests.solvers.gurobi_to_pyomo_expressions import (
    grb_nl_to_pyo_expr,
)
from pyomo.core.expr.compare import assertExpressionsEqual
from pyomo.core.expr.numeric_expr import SumExpression, ProductExpression
from pyomo.environ import (
    Binary,
    BooleanVar,
    ConcreteModel,
    Constraint,
    Expression,
    Integers,
    log,
    LogicalConstraint,
    maximize,
    NonNegativeIntegers,
    NonNegativeReals,
    NonPositiveIntegers,
    NonPositiveReals,
    Objective,
    Param,
    Reals,
    value,
    Var,
)
from pyomo.gdp import Disjunction
from pyomo.opt import WriterFactory
from pyomo.contrib.solver.solvers.gurobi.gurobi_direct_minlp import (
    GurobiDirectMINLP,
    GurobiMINLPVisitor,
)
from pyomo.contrib.solver.common.factory import SolverFactory
from pyomo.contrib.solver.common.results import TerminationCondition
from pyomo.contrib.solver.tests.solvers.test_gurobi_minlp_walker import CommonTest

gurobipy, gurobipy_available = attempt_import('gurobipy', minimum_version='12.0.0')
if gurobipy_available:
    from gurobipy import GRB

    if not GurobiDirectMINLP().available():
        gurobipy_available = False


def make_model():
    m = ConcreteModel()
    m.x1 = Var(domain=NonNegativeReals, bounds=(0, 10))
    m.x2 = Var(domain=Reals, bounds=(-3, 4))
    m.x3 = Var(domain=NonPositiveReals, bounds=(-13, 0))
    m.y1 = Var(domain=Integers, bounds=(4, 14))
    m.y2 = Var(domain=NonNegativeIntegers, bounds=(5, 16))
    m.y3 = Var(domain=NonPositiveIntegers, bounds=(-13, 0))
    m.z1 = Var(domain=Binary)

    m.c1 = Constraint(expr=2**m.x2 >= m.x3)
    m.c2 = Constraint(expr=m.y1**2 <= 7)
    m.c3 = Constraint(expr=m.y2 + m.y3 + 5 * m.z1 >= 17)

    m.obj = Objective(expr=log(m.x1))

    return m


@unittest.skipUnless(gurobipy_available, "Gurobipy 12 is not available")
@unittest.pytest.mark.solver("gurobi_direct_minlp")
class TestGurobiMINLPWriter(CommonTest):
    def test_small_model(self):
        grb_model = gurobipy.Model()
        visitor = GurobiMINLPVisitor(grb_model, symbolic_solver_labels=True)

        m = make_model()

        grb_model, var_map, obj, grb_cons, pyo_cons = WriterFactory(
            'gurobi_minlp'
        ).write(m, symbolic_solver_labels=True)

        self.assertEqual(len(var_map), 7)
        x1 = var_map[id(m.x1)]
        x2 = var_map[id(m.x2)]
        x3 = var_map[id(m.x3)]
        y1 = var_map[id(m.y1)]
        y2 = var_map[id(m.y2)]
        y3 = var_map[id(m.y3)]
        z1 = var_map[id(m.z1)]

        self.assertEqual(grb_model.numVars, 9)
        self.assertEqual(grb_model.numIntVars, 4)
        self.assertEqual(grb_model.numBinVars, 1)

        lin_constrs = grb_model.getConstrs()
        self.assertEqual(len(lin_constrs), 2)
        quad_constrs = grb_model.getQConstrs()
        self.assertEqual(len(quad_constrs), 1)
        nonlinear_constrs = grb_model.getGenConstrs()
        self.assertEqual(len(nonlinear_constrs), 2)

        ## linear constraints

        # this is the linear piece of c1
        c = lin_constrs[0]
        c_expr = grb_model.getRow(c)
        self.assertEqual(c.RHS, 0)
        self.assertEqual(c.Sense, '<')
        self.assertEqual(c_expr.size(), 1)
        self.assertEqual(c_expr.getCoeff(0), 1)
        self.assertEqual(c_expr.getConstant(), 0)
        aux_var = c_expr.getVar(0)

        c3 = lin_constrs[1]
        c3_expr = grb_model.getRow(c3)
        self.assertEqual(c3_expr.size(), 3)
        self.assertIs(c3_expr.getVar(0), y2)
        self.assertEqual(c3_expr.getCoeff(0), 1)
        self.assertIs(c3_expr.getVar(1), y3)
        self.assertEqual(c3_expr.getCoeff(1), 1)
        self.assertIs(c3_expr.getVar(2), z1)
        self.assertEqual(c3_expr.getCoeff(2), 5)
        self.assertEqual(c3_expr.getConstant(), 0)
        self.assertEqual(c3.RHS, 17)
        self.assertEqual(c3.Sense, '>')

        ## quadratic constraint
        c2 = quad_constrs[0]
        c2_expr = grb_model.getQCRow(c2)
        lin_expr = c2_expr.getLinExpr()
        self.assertEqual(lin_expr.size(), 0)
        self.assertEqual(lin_expr.getConstant(), 0)
        self.assertEqual(c2.QCRHS, 7)
        self.assertEqual(c2.QCSense, '<')
        self.assertEqual(c2_expr.size(), 1)
        self.assertIs(c2_expr.getVar1(0), y1)
        self.assertIs(c2_expr.getVar2(0), y1)
        self.assertEqual(c2_expr.getCoeff(0), 1)

        ## general nonlinear constraints
        obj_cons = nonlinear_constrs[0]
        res_var, opcode, data, parent = grb_model.getGenConstrNLAdv(obj_cons)
        self.assertEqual(len(opcode), 2)  # two nodes in the expression tree
        self.assertEqual(opcode[0], GRB.OPCODE_LOG)
        # log has no data
        self.assertEqual(parent[0], -1)  # it's the root
        self.assertEqual(opcode[1], GRB.OPCODE_VARIABLE)
        self.assertIs(data[1], x1)
        self.assertEqual(parent[1], 0)

        # we can check that res_var is the objective
        self.assertEqual(grb_model.ModelSense, 1)  # minimizing
        obj = grb_model.getObjective()
        self.assertEqual(obj.size(), 1)
        self.assertEqual(obj.getCoeff(0), 1)
        self.assertIs(obj.getVar(0), res_var)

        c1 = nonlinear_constrs[1]
        res_var, opcode, data, parent = grb_model.getGenConstrNLAdv(c1)
        # This is where we link into the linear inequality constraint
        self.assertIs(res_var, aux_var)
        # test the tree for the expression x3  + (- (2 ** x2))
        self.assertEqual(len(opcode), 6)
        self.assertEqual(opcode[0], GRB.OPCODE_PLUS)
        # plus has no data
        self.assertEqual(parent[0], -1)  # root
        self.assertEqual(opcode[1], GRB.OPCODE_VARIABLE)
        self.assertIs(data[1], x3)
        self.assertEqual(parent[1], 0)
        self.assertEqual(opcode[2], GRB.OPCODE_UMINUS)  # negation
        # negation has no data
        self.assertEqual(parent[2], 0)
        self.assertEqual(opcode[3], GRB.OPCODE_POW)
        # pow has no data
        self.assertEqual(parent[3], 2)
        self.assertEqual(opcode[4], GRB.OPCODE_CONSTANT)
        self.assertEqual(data[4], 2)
        self.assertEqual(parent[4], 3)
        self.assertEqual(opcode[5], GRB.OPCODE_VARIABLE)
        self.assertIs(data[5], x2)
        self.assertEqual(parent[5], 3)

    def test_write_NPV_negation_in_RHS(self):
        m = ConcreteModel()
        m.x1 = Var()
        m.p1 = Param(initialize=3, mutable=True)
        m.c = Constraint(expr=-m.x1 == m.p1)
        m.obj = Objective(expr=m.x1)

        grb_model, var_map, obj, grb_cons, pyo_cons = WriterFactory(
            'gurobi_minlp'
        ).write(m, symbolic_solver_labels=True)

        self.assertEqual(len(var_map), 1)
        x1 = var_map[id(m.x1)]

        self.assertEqual(grb_model.numVars, 1)
        self.assertEqual(grb_model.numIntVars, 0)
        self.assertEqual(grb_model.numBinVars, 0)

        lin_constrs = grb_model.getConstrs()
        self.assertEqual(len(lin_constrs), 1)
        quad_constrs = grb_model.getQConstrs()
        self.assertEqual(len(quad_constrs), 0)
        nonlinear_constrs = grb_model.getGenConstrs()
        self.assertEqual(len(nonlinear_constrs), 0)

        # constraint
        c = lin_constrs[0]
        c_expr = grb_model.getRow(c)
        self.assertEqual(c.RHS, 3)
        self.assertEqual(c.Sense, '=')
        self.assertEqual(c_expr.size(), 1)
        self.assertEqual(c_expr.getCoeff(0), -1)
        self.assertEqual(c_expr.getConstant(), 0)
        self.assertIs(c_expr.getVar(0), x1)

        # objective
        self.assertEqual(grb_model.ModelSense, 1)  # minimizing
        obj = grb_model.getObjective()
        self.assertEqual(obj.size(), 1)
        self.assertEqual(obj.getCoeff(0), 1)
        self.assertIs(obj.getVar(0), x1)

    def test_writer_ignores_deactivated_logical_constraints(self):
        m = ConcreteModel()
        m.x1 = Var()
        m.p1 = Param(initialize=3, mutable=True)
        m.c = Constraint(expr=-m.x1 == m.p1)
        m.obj = Objective(expr=m.x1)

        m.b = BooleanVar()
        m.whatever = LogicalConstraint(expr=~m.b)
        m.whatever.deactivate()

        grb_model, var_map, obj, grb_cons, pyo_cons = WriterFactory(
            'gurobi_minlp'
        ).write(m, symbolic_solver_labels=True)

        self.assertEqual(len(var_map), 1)
        x1 = var_map[id(m.x1)]

        self.assertEqual(grb_model.numVars, 1)
        self.assertEqual(grb_model.numIntVars, 0)
        self.assertEqual(grb_model.numBinVars, 0)

        lin_constrs = grb_model.getConstrs()
        self.assertEqual(len(lin_constrs), 1)
        quad_constrs = grb_model.getQConstrs()
        self.assertEqual(len(quad_constrs), 0)
        nonlinear_constrs = grb_model.getGenConstrs()
        self.assertEqual(len(nonlinear_constrs), 0)

        # constraint
        c = lin_constrs[0]
        c_expr = grb_model.getRow(c)
        self.assertEqual(c.RHS, 3)
        self.assertEqual(c.Sense, '=')
        self.assertEqual(c_expr.size(), 1)
        self.assertEqual(c_expr.getCoeff(0), -1)
        self.assertEqual(c_expr.getConstant(), 0)
        self.assertIs(c_expr.getVar(0), x1)

        # objective
        self.assertEqual(grb_model.ModelSense, 1)  # minimizing
        obj = grb_model.getObjective()
        self.assertEqual(obj.size(), 1)
        self.assertEqual(obj.getCoeff(0), 1)
        self.assertIs(obj.getVar(0), x1)

    def test_named_expression_quadratic(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.e = Expression(expr=m.x**2 + m.y)
        m.c = Constraint(expr=m.e <= 7)
        m.c2 = Constraint(expr=m.e >= -3)
        m.obj = Objective(expr=0)

        grb_model, var_map, obj, grb_cons, pyo_cons = WriterFactory(
            'gurobi_minlp'
        ).write(m, symbolic_solver_labels=True)

        self.assertEqual(len(var_map), 2)
        x = var_map[id(m.x)]
        y = var_map[id(m.y)]

        self.assertEqual(grb_model.numVars, 2)
        self.assertEqual(grb_model.numIntVars, 0)
        self.assertEqual(grb_model.numBinVars, 0)

        lin_constrs = grb_model.getConstrs()
        self.assertEqual(len(lin_constrs), 0)
        quad_constrs = grb_model.getQConstrs()
        self.assertEqual(len(quad_constrs), 2)
        nonlinear_constrs = grb_model.getGenConstrs()
        self.assertEqual(len(nonlinear_constrs), 0)

        # constraint 1
        c1 = quad_constrs[0]
        expr = grb_model.getQCRow(c1)
        lin_expr = expr.getLinExpr()
        self.assertEqual(lin_expr.size(), 1)
        self.assertEqual(lin_expr.getConstant(), 0)
        self.assertEqual(lin_expr.getCoeff(0), 1)
        self.assertIs(lin_expr.getVar(0), y)
        self.assertEqual(c1.QCRHS, 7)
        self.assertEqual(c1.QCSense, '<')
        self.assertEqual(expr.size(), 1)
        self.assertIs(expr.getVar1(0), x)
        self.assertIs(expr.getVar2(0), x)
        self.assertEqual(expr.getCoeff(0), 1)

        # constraint 2
        c2 = quad_constrs[1]
        expr = grb_model.getQCRow(c1)
        lin_expr = expr.getLinExpr()
        self.assertEqual(lin_expr.size(), 1)
        self.assertEqual(lin_expr.getConstant(), 0)
        self.assertEqual(lin_expr.getCoeff(0), 1)
        self.assertIs(lin_expr.getVar(0), y)
        self.assertEqual(c2.QCRHS, -3)
        self.assertEqual(c2.QCSense, '>')
        self.assertEqual(expr.size(), 1)
        self.assertIs(expr.getVar1(0), x)
        self.assertIs(expr.getVar2(0), x)
        self.assertEqual(expr.getCoeff(0), 1)

        # objective
        self.assertEqual(grb_model.ModelSense, 1)  # minimizing
        obj = grb_model.getObjective()
        self.assertEqual(obj.size(), 0)

    def test_named_expression_nonlinear(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var()
        m.e = Expression(expr=log(m.x) ** 2 + m.y)
        m.c = Constraint(expr=m.e <= 7)
        m.c2 = Constraint(expr=m.e + m.y**3 + log(m.x + m.y) >= -3)
        m.obj = Objective(expr=0)

        grb_model, var_map, obj, grb_cons, pyo_cons = WriterFactory(
            'gurobi_minlp'
        ).write(m, symbolic_solver_labels=True)

        self.assertEqual(len(var_map), 2)
        x = var_map[id(m.x)]
        y = var_map[id(m.y)]
        reverse_var_map = {grbv: pyov for pyov, grbv in var_map.items()}

        self.assertEqual(grb_model.numVars, 4)
        self.assertEqual(grb_model.numIntVars, 0)
        self.assertEqual(grb_model.numBinVars, 0)

        lin_constrs = grb_model.getConstrs()
        self.assertEqual(len(lin_constrs), 2)
        quad_constrs = grb_model.getQConstrs()
        self.assertEqual(len(quad_constrs), 0)
        nonlinear_constrs = grb_model.getGenConstrs()
        self.assertEqual(len(nonlinear_constrs), 2)

        c1 = lin_constrs[0]
        aux1 = grb_model.getRow(c1)
        self.assertEqual(c1.RHS, 7)
        self.assertEqual(c1.Sense, '<')
        self.assertEqual(aux1.size(), 1)
        self.assertEqual(aux1.getCoeff(0), 1)
        self.assertEqual(aux1.getConstant(), 0)
        aux1 = aux1.getVar(0)

        c2 = lin_constrs[1]
        aux2 = grb_model.getRow(c2)
        self.assertEqual(c2.RHS, -3)
        self.assertEqual(c2.Sense, '>')
        self.assertEqual(aux2.size(), 1)
        self.assertEqual(aux2.getCoeff(0), 1)
        self.assertEqual(aux2.getConstant(), 0)
        aux2 = aux2.getVar(0)

        # log(x)**2 + y
        g1 = nonlinear_constrs[0]
        aux_var, opcode, data, parent = grb_model.getGenConstrNLAdv(g1)
        self.assertIs(aux_var, aux1)
        assertExpressionsEqual(
            self,
            grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map),
            log(m.x) ** 2 + m.y,
        )

        # log(x)**2 + y + y**3 + log(x + y)
        g2 = nonlinear_constrs[1]
        aux_var, opcode, data, parent = grb_model.getGenConstrNLAdv(g2)
        self.assertIs(aux_var, aux2)
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)
        assertExpressionsEqual(
            self,
            pyo_expr,
            SumExpression(
                (
                    SumExpression((SumExpression((log(m.x) ** 2, m.y)), m.y**3.0)),
                    log(SumExpression((m.x, m.y))),
                )
            ),
        )

        # objective
        self.assertEqual(grb_model.ModelSense, 1)  # minimizing
        obj = grb_model.getObjective()
        self.assertEqual(obj.size(), 0)

    def test_solve_model(self):
        m = ConcreteModel()
        m.x = Var(bounds=(0, 1))
        m.y = Var()
        m.c = Constraint(expr=m.y == m.x**2)
        m.obj = Objective(expr=m.x + m.y, sense=maximize)

        results = SolverFactory('gurobi_direct_minlp').solve(m, tee=True)

        self.assertEqual(value(m.obj.expr), 2)

        self.assertEqual(value(m.x), 1)
        self.assertEqual(value(m.y), 1)

        self.assertEqual(
            results.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(results.incumbent_objective, 2)
        self.assertEqual(results.objective_bound, 2)

    def test_unbounded_because_of_multiplying_by_0(self):
        # Gurobi believes that the expression in m.c is nonlinear, so we have
        # to pass it that way for this to work. Because this is in fact an
        # unbounded model.

        m = ConcreteModel()
        m.x1 = Var()
        m.x2 = Var()
        m.x3 = Var()
        m.c = Constraint(expr=(0 * m.x1 * m.x2) * m.x3 == 0)
        m.obj = Objective(expr=m.x1)

        grb_model, var_map, obj, grb_cons, pyo_cons = WriterFactory(
            'gurobi_minlp'
        ).write(m, symbolic_solver_labels=True)

        self.assertEqual(len(var_map), 3)
        x1 = var_map[id(m.x1)]
        x2 = var_map[id(m.x2)]
        x3 = var_map[id(m.x3)]

        self.assertEqual(grb_model.numVars, 4)
        self.assertEqual(grb_model.numIntVars, 0)
        self.assertEqual(grb_model.numBinVars, 0)

        lin_constrs = grb_model.getConstrs()
        self.assertEqual(len(lin_constrs), 1)
        quad_constrs = grb_model.getQConstrs()
        self.assertEqual(len(quad_constrs), 0)
        nonlinear_constrs = grb_model.getGenConstrs()
        self.assertEqual(len(nonlinear_constrs), 1)

        # this is the auxiliary variable equality
        c = lin_constrs[0]
        c_expr = grb_model.getRow(c)
        self.assertEqual(c.RHS, 0)
        self.assertEqual(c.Sense, '=')
        self.assertEqual(c_expr.size(), 1)
        self.assertEqual(c_expr.getCoeff(0), 1)
        self.assertEqual(c_expr.getConstant(), 0)
        aux_var = c_expr.getVar(0)

        # this is the nonlinear equality
        c = nonlinear_constrs[0]
        res_var, opcode, data, parent = grb_model.getGenConstrNLAdv(c)
        # This is where we link into the linear inequality constraint
        self.assertIs(res_var, aux_var)
        reverse_var_map = {grb_v: pyo_v for pyo_v, grb_v in var_map.items()}
        pyo_expr = grb_nl_to_pyo_expr(opcode, data, parent, reverse_var_map)

        assertExpressionsEqual(
            self,
            pyo_expr,
            ProductExpression((ProductExpression((0.0, m.x1, m.x2, m.x3)),)),
        )

        opt = SolverFactory('gurobi_direct_minlp')
        opt.config.raise_exception_on_nonoptimal_result = False
        results = opt.solve(m)
        # model is unbounded
        self.assertEqual(results.termination_condition, TerminationCondition.unbounded)

    @unittest.skipUnless(numpy_available, "Numpy is not available")
    def test_numpy_trivially_true_constraint(self):
        m = ConcreteModel()
        m.x1 = Var()
        m.x2 = Var()
        m.x1.fix(np.float64(0))
        m.x2.fix(np.float64(0))
        m.c = Constraint(expr=m.x1 == m.x2)
        m.obj = Objective(expr=m.x1)
        results = SolverFactory('gurobi_direct_minlp').solve(m)

        self.assertEqual(
            results.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(value(m.obj), 0)
        self.assertEqual(results.incumbent_objective, 0)
        self.assertEqual(results.objective_bound, 0)

    def test_trivially_true_constraint(self):
        """
        We can pass trivially true things to Gurobi and it's fine
        """
        m = ConcreteModel()
        m.x1 = Var()
        m.x2 = Var()
        m.x1.fix(2)
        m.x2.fix(2)
        m.c = Constraint(expr=m.x1 <= m.x2)
        m.obj = Objective(expr=m.x1)
        results = SolverFactory('gurobi_direct_minlp').solve(m, tee=True)

        self.assertEqual(
            results.termination_condition,
            TerminationCondition.convergenceCriteriaSatisfied,
        )
        self.assertEqual(value(m.obj), 2)
        self.assertEqual(results.incumbent_objective, 2)
        self.assertEqual(results.objective_bound, 2)

    def test_multiple_objective_error(self):
        m = make_model()
        m.obj2 = Objective(expr=m.x1 + m.x2)

        with self.assertRaisesRegex(
            ValueError,
            "More than one active objective defined for input model 'unknown': "
            "Cannot write to gurobipy",
        ):
            results = SolverFactory('gurobi_direct_minlp').solve(m)

    def test_unrecognized_component_error(self):
        m = make_model()
        m.disj = Disjunction(expr=[[m.x1 + m.x2 == 3], [m.x1 + m.x2 >= 7]])

        with self.assertRaisesRegex(
            ValueError,
            r"The model \('unknown'\) contains the following active components "
            r"that the Gurobi MINLP writer does not know how to process:"
            + "\n\t"
            + r"\<class 'pyomo.gdp.disjunct.Disjunct'\>:"
            + "\n\t\t"
            + r"disj_disjuncts\[0\]"
            + "\n\t\t"
            + r"disj_disjuncts\[1\]"
            + "\n\t"
            + r"<class 'pyomo.gdp.disjunct.Disjunction'>:"
            + "\n\t\t"
            + "disj",
        ):
            results = SolverFactory('gurobi_direct_minlp').solve(m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import copy
import weakref
from pyomo.common.autoslots import AutoSlots


def _not_implemented(*args, **kwds):
    raise NotImplementedError("This property is abstract")  # pragma:nocover


def _abstract_readwrite_property(**kwds):
    p = property(fget=_not_implemented, fset=_not_implemented, **kwds)
    return p


def _abstract_readonly_property(**kwds):
    p = property(fget=_not_implemented, **kwds)
    return p


class _no_ctype:
    """The default argument for methods that accept a ctype."""

    pass


# This will be populated outside of core.kernel. It will map
# AML classes (which are the ctypes used by all of the
# solver interfaces) to Kernel classes
_convert_ctype = {}
_kernel_ctype_backmap = {}


def _convert_descend_into(value):
    """Converts the descend_into keyword to a function"""
    if hasattr(value, "__call__"):
        return value
    elif value:
        return _convert_descend_into._true
    else:
        return _convert_descend_into._false


_convert_descend_into._true = lambda x: True
_convert_descend_into._false = lambda x: False


class ICategorizedObject(AutoSlots.Mixin):
    """
    Interface for objects that maintain a weak reference to
    a parent storage object and have a category type.

    This class is abstract. It assumes any derived class
    declares the attributes below with or without slots:

    Attributes:
        _ctype: Stores the object's category type, which
            should be some class derived from
            ICategorizedObject. This attribute may be
            declared at the class level.
        _parent: Stores a weak reference to the object's
            parent container or :const:`None`.
        _storage_key: Stores key this object can be accessed
            with through its parent container.
        _active (bool): Stores the active status of this
            object.
    """

    __slots__ = ()
    __autoslot_mappers__ = {'_parent': AutoSlots.weakref_mapper}

    # These flags can be used by implementations
    # to avoid isinstance calls.
    _is_container = False
    """A flag used to indicate that the class is an instance
    of ICategorizedObjectContainer."""

    _is_heterogeneous_container = False
    """A flag used to indicate that the class is an instance
    of ICategorizedObjectContainer that stores objects with
    different category types than its own."""

    #
    # Interface
    #

    @property
    def ctype(self):
        """The object's category type."""
        return self._ctype

    @property
    def parent(self):
        """The object's parent (possibly None)."""
        if isinstance(self._parent, weakref.ReferenceType):
            return self._parent()
        else:
            return self._parent

    @property
    def storage_key(self):
        """The object's storage key within its parent"""
        return self._storage_key

    @property
    def active(self):
        """The active status of this object."""
        return self._active

    @active.setter
    def active(self, value):
        raise AttributeError("Assignment not allowed. Use the (de)activate method")

    ### The following group of methods use object.__setattr__
    ### to update the _parent, _storage_key, and _active flags.
    ### This is done to allow the block implementation (block.py)
    ### to protect them from being overwritten with user added
    ### components in its __setattr__ method
    def _update_parent_and_storage_key(self, parent, key):
        object.__setattr__(self, "_parent", weakref.ref(parent))
        object.__setattr__(self, "_storage_key", key)

    def _clear_parent_and_storage_key(self):
        object.__setattr__(self, "_parent", None)
        object.__setattr__(self, "_storage_key", None)

    def activate(self):
        """Activate this object."""
        object.__setattr__(self, "_active", True)

    def deactivate(self):
        """Deactivate this object."""
        object.__setattr__(self, "_active", False)

    ###

    def getname(
        self,
        fully_qualified=False,
        name_buffer={},  # HACK: ignored (required to work with some solver interfaces, but that code should change soon)
        convert=str,
        relative_to=None,
    ):
        """
        Dynamically generates a name for this object.

        Args:
            fully_qualified (bool): Generate a full name by
                iterating through all ancestor containers.
                Default is :const:`False`.
            convert (function): A function that converts a
                storage key into a string
                representation. Default is the built-in
                function str.
            relative_to : When generating a fully
                qualified name, generate the name relative
                to this block.

        Returns:
            If a parent exists, this method returns a string
            representing the name of the object in the
            context of its parent; otherwise (if no parent
            exists), this method returns :const:`None`.
        """
        assert fully_qualified or (relative_to is None)
        parent = self.parent
        if parent is None:
            return None

        key = self.storage_key
        name = parent._child_storage_entry_string % convert(key)
        if fully_qualified:
            parent_name = parent.getname(fully_qualified=True, relative_to=relative_to)
            if (parent_name is not None) and (
                (relative_to is None) or (parent is not relative_to)
            ):
                return parent_name + parent._child_storage_delimiter_string + name
            else:
                return name
        else:
            return name

    @property
    def name(self):
        """The object's fully qualified name. Alias for
        `obj.getname(fully_qualified=True)`."""
        return self.getname(fully_qualified=True)

    @property
    def local_name(self):
        """The object's local name within the context of its
        parent. Alias for
        `obj.getname(fully_qualified=False)`."""
        return self.getname(fully_qualified=False)

    def __str__(self):
        """Convert this object to a string by first
        attempting to generate its fully qualified name. If
        the object does not have a name (because it does not
        have a parent, then a string containing the class
        name is returned."""
        name = self.name
        if name is None:
            return "<" + self.__class__.__name__ + ">"
        else:
            return name

    def clone(self):
        """
        Returns a copy of this object with the parent
        pointer set to :const:`None`.

        A clone is almost equivalent to deepcopy except that
        any categorized objects encountered that are not
        descendents of this object will reference the same
        object on the clone.
        """
        save_parent = self._parent
        object.__setattr__(self, "_parent", None)
        try:
            new_block = copy.deepcopy(
                self, {'__block_scope__': {id(self): True, id(None): False}}
            )
        finally:
            object.__setattr__(self, "_parent", save_parent)
        return new_block

    #
    # The following method must be defined to allow
    # expressions and blocks to be correctly cloned.
    # (copied from JDS code in original component.py,
    # comments removed for now)
    #

    def __deepcopy__(self, memo):
        if '__block_scope__' in memo:
            _known = memo['__block_scope__']
            _new = None
            tmp = self.parent
            _in_scope = tmp is None
            while id(tmp) not in _known:
                _new = (_new, id(tmp))
                tmp = tmp.parent
            _in_scope |= _known[id(tmp)]

            while _new is not None:
                _new, _id = _new
                _known[_id] = _in_scope

            if not _in_scope and id(self) not in _known:
                # component is out-of-scope. shallow copy only
                memo[id(self)] = self
                return self

        if id(self) in memo:
            return memo[id(self)]

        ans = memo[id(self)] = self.__class__.__new__(self.__class__)
        ans.__setstate__(copy.deepcopy(self.__getstate__(), memo))
        return ans


class ICategorizedObjectContainer(ICategorizedObject):
    """
    Interface for categorized containers of categorized
    objects.
    """

    _is_container = True
    _child_storage_delimiter_string = None
    _child_storage_entry_string = None
    __slots__ = ()

    #
    # Interface
    #

    def activate(self, shallow=True):
        """Activate this container."""
        super(ICategorizedObjectContainer, self).activate()
        if not shallow:
            for child in self.children():
                if not child._is_container:
                    child.activate()
                else:
                    child.activate(shallow=False)

    def deactivate(self, shallow=True):
        """Deactivate this container."""
        super(ICategorizedObjectContainer, self).deactivate()
        if not shallow:
            for child in self.children():
                if not child._is_container:
                    child.deactivate()
                else:
                    child.deactivate(shallow=False)

    def child(self, *args, **kwds):
        """Returns a child of this container given a storage
        key."""
        raise NotImplementedError  # pragma:nocover

    def children(self, *args, **kwds):
        """A generator over the children of this container."""
        raise NotImplementedError  # pragma:nocover

    def components(self, *args, **kwds):
        """A generator over the set of components stored
        under this container."""
        raise NotImplementedError  # pragma:nocover
